//  Created by T Brennan on 30/1/21.

import 'dart:async';
import 'dart:math';

import 'package:flutter/cupertino.dart';
import 'package:get_it/get_it.dart';
import 'package:memri/core/cvu/services/resolving/cvu_context.dart';
import 'package:memri/core/cvu/services/resolving/cvu_property_resolver.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/cvu/controllers/cvu_controller.dart';
import 'package:memri/core/cvu/controllers/cvu_lookup_controller.dart';
import 'package:memri/core/cvu/models/cvu_parsed_definition.dart';
import 'package:memri/core/cvu/models/cvu_value.dart';
import 'package:memri/core/cvu/models/cvu_view_arguments.dart';
import 'package:memri/core/cvu/models/view_context.dart';
import 'package:memri/core/cvu/utilities/binding.dart';
import 'package:memri/utilities/extensions/collection.dart';
import 'package:memri/utilities/helpers/app_helper.dart';

import 'database_query.dart';

class ViewContextController extends ChangeNotifier {
  late ViewContextHolder configHolder;

  late CVUController cvuController;
  late CVULookupController lookupController;

  StreamSubscription? queryObservation;

  CVUDefinitionContent? viewDefinition;
  CVUDefinitionContent rendererDefinition;

  ViewContext get config => configHolder.config;

  ViewContextController(
      {required ViewContextHolder config,
      CVUController? cvuController,
      rendererDefinition})
      : rendererDefinition = rendererDefinition ?? CVUDefinitionContent() {
    this.cvuController = cvuController ?? GetIt.instance();
    lookupController = CVULookupController();
    config.config.viewArguments ?? CVUViewArguments();
    configHolder = config;

    if (this.config.viewName != null && this.config.viewName != 'customView') {
      viewDefinition = config.config.viewDefinition;
    }

    _updateCachedValues();

    setupQueryObservation();
    this.cvuController.addListener(() => updateUI());
  }

  factory ViewContextController.fromParams({
    String viewName = 'customView',
    String rendererName = 'custom',
    CVUDefinitionContent? viewDefinition,
    CVUViewArguments? viewArguments,
    Item? focusedItem,
    CVUController? cvuController,
  }) {
    viewArguments ??= CVUViewArguments();
    viewArguments.argumentItem = focusedItem;
    cvuController ??= GetIt.instance();

    viewDefinition ??= cvuController.viewDefinitionFor(viewName: viewName) ??
        CVUDefinitionContent();
    var newContext = CVUContext(
        currentItem: focusedItem,
        selector: null,
        viewName: viewName,
        viewDefinition: viewDefinition,
        viewArguments: viewArguments);
    var datasource = viewDefinition.definitions.firstWhereOrNull(
        (definition) => definition.type == CVUDefinitionType.datasource);
    var queryConfig = DatabaseQueryConfig.queryConfigWith(
        context: newContext, datasource: datasource);

    var defaultRenderer = 'list';
    var rendererName = ((() =>
        viewDefinition!
            .propertyResolver(
              context: newContext,
              lookup: CVULookupController(),
            )
            .string('defaultRenderer') ??
        defaultRenderer))();

    var config = ViewContext(
        viewName: viewName,
        rendererName: rendererName,
        viewDefinition: viewDefinition,
        viewArguments: viewArguments,
        focusedItem: focusedItem,
        query: queryConfig);

    var holder = ViewContextHolder(config);
    return ViewContextController(config: holder, cvuController: cvuController);
  }

  int get focusedIndex {
    var focusedItem = config.focusedItem;
    if (focusedItem == null) {
      return 0;
    }
    return max(items.indexOf(focusedItem), 0);
  }

  set focusedIndex(int newValue) {
    var item = items.asMap()[newValue];
    if (item != null) {
      config.focusedItem = item;
    } else {
      config.focusedItem = null;
    }
  }

  Item? get focusedItem => config.focusedItem;

  set focusedItem(Item? newValue) {
    config.focusedItem = newValue;
  }

  onAppear() {
    isObservingQuery = true;

    configHolder.addListener(_updateCachedValues);
  }

  onDisappear() {
    isObservingQuery = false;

    configHolder.removeListener(_updateCachedValues);
  }

  /// Return a Widget for the given item based on it's CVU definition.
  /// Set `overrideRenderer` if you want to render the item as though it is in a different renderer to the context (eg. "list" to get the list-specific appearance)
  Widget render(
      {Item? item,
      List<Item>? items,
      String? overrideRenderer,
      CVUDefinitionContent? overrideViewDefinition,
      CVUDefinitionContent? nodeDefinition,
      CVUViewArguments? viewArguments,
      bool blankIfNoDefinition = false}) {
    var key = item != null ? Key(item.id) : null;
    return cvuController.render(
        cvuContext: getCVUContext(
            item: item,
            items: items,
            overrideViewDefinition: overrideViewDefinition,
            overrideRenderer: overrideRenderer,
            viewArguments: viewArguments),
        nodeDefinition: nodeDefinition,
        lookup: lookupController,
        blankIfNoDefinition: blankIfNoDefinition,
        key: key);
  }

  CVUContext getCVUContext(
      {Item? item,
      List<Item>? items,
      CVUViewArguments? viewArguments,
      CVUDefinitionContent? overrideViewDefinition,
      String? overrideRenderer}) {
    return CVUContext(
        currentItem: item,
        items: items,
        selector: null,
        viewName: config.viewName,
        rendererName: overrideRenderer ?? config.rendererName,
        viewDefinition: overrideViewDefinition ?? config.viewDefinition,
        viewArguments: viewArguments ?? config.viewArguments);
  }

  _updateCachedValues() {
    var viewName = config.viewName;
    viewDefinition ??= viewName != null
        ? cvuController.viewDefinitionFor(viewName: viewName)
        : null; //TODO this part is full of unused legacy we keep stumbling on

    if (viewDefinition == null) {
      var item = config.focusedItem;
      viewDefinition = item != null
          ? cvuController.viewDefinitionForItemRecord(item: item)
          : null;
    }

    if (viewDefinition != null) {
      config.viewDefinition = viewDefinition!;

      var datasource = config.viewDefinition.definitions.firstWhereOrNull(
          (definition) => definition.type == CVUDefinitionType.datasource);
      var queryConfig = DatabaseQueryConfig.queryConfigWith(
          context: getCVUContext(item: config.focusedItem),
          datasource: datasource);
      config.query.queryGraphQL = queryConfig.queryGraphQL;
    }

    rendererDefinition =
        cvuController.rendererDefinitionFor(getCVUContext())?.parsed ??
            CVUDefinitionContent();

    viewDefinitionPropertyResolver = config.viewDefinition.propertyResolver(
      context: getCVUContext(),
      lookup: CVULookupController(),
    );

    rendererDefinitionPropertyResolver = rendererDefinition.propertyResolver(
      context: getCVUContext(),
      lookup: CVULookupController(),
    );
    notifyListeners();
  }

  Set<String> get supportedRenderers {
    var defined =
        (viewDefinitionPropertyResolver.stringArray('supportedRenderers'))
            .map((property) => property.toLowerCase());
    if (defined.isNotEmpty) {
      return Set.of(defined);
    } else {
      var defaultRenderer =
          (viewDefinitionPropertyResolver.string('defaultRenderer'))
              ?.toLowerCase();
      List<String> list = config.viewDefinition.definitions
          .map((def) {
            if (def.type != CVUDefinitionType.renderer) {
              return null;
            }
            return def.name?.toLowerCase();
          })
          .whereType<String>()
          .toList();
      if (defaultRenderer != null) {
        list.add(defaultRenderer);
      }
      return Set.of(list);
    }
  }

  late CVUPropertyResolver viewDefinitionPropertyResolver = () {
    return config.viewDefinition.propertyResolver(
      context: getCVUContext(),
      lookup: CVULookupController(),
    );
  }();

  late CVUPropertyResolver rendererDefinitionPropertyResolver = () {
    return rendererDefinition.propertyResolver(
      context: getCVUContext(),
      lookup: CVULookupController(),
    );
  }();

  CVUPropertyResolver? get itemPropertyResolver {
    var viewDefinition =
        cvuController.viewDefinitionForItemRecord(item: focusedItem);
    if (viewDefinition == null) {
      return null;
    }
    return viewDefinition.propertyResolver(
      context: getCVUContext(item: focusedItem),
      lookup: CVULookupController(),
    );
  }

  setRendererProperty(String renderer, String property, CVUValue value) {
    var index = config.viewDefinition.definitions.indexWhere((definition) =>
        definition.type == CVUDefinitionType.renderer &&
        definition.name == renderer);
    if (index >= 0) {
      config.viewDefinition.definitions[index].parsed.properties[property] =
          value;
    } else {
      var props = CVUDefinitionContent(properties: {property: value});
      var def = CVUParsedDefinition(
          type: CVUDefinitionType.renderer, name: renderer, parsed: props);
      config.viewDefinition.definitions.add(def);
    }
    _updateCachedValues();
  }

  CVUPropertyResolver? nodePropertyResolver(Item item) {
    var context = getCVUContext(item: item);
    return cvuController.nodeDefinitionFor(context)?.propertyResolver(
          context: context,
          lookup: CVULookupController(),
        );
  }

  // MARK: Query RESULT
  /// This holds the array of results for the query. If `isObserving` is set to true, this will be up to date with any changes in the database.
  List<Item>? _items;

  List<Item> get items => _items ?? [];

  set items(List<Item> items) {
    _items = items;
    // itemsValueNotifier.value = _itemRecords!;
    notifyListeners();
  }

  ValueNotifier<List<Item>> itemsValueNotifier =
      ValueNotifier([]); //TODO not sure if this is right architecture

  /// Convenience variable returning whether the items are loaded or not
  bool get isLoaded => _items != null;

  /// Convenience variable returning whether there are any items
  bool get hasItems => items.isNotEmpty || items.isNotEmpty;

  // MARK: Selection State
  List<String> _selectedItems = <String>[];

  get selectedItems => _selectedItems;

  set selectedItems(selectedItems) {
    _selectedItems = selectedItems;
    config.viewArguments?.args['selectedItems'] =
        CVUValueArray(_selectedItems.compactMap((uid) => CVUValueItem(uid)));
  }

  Binding<Set<int>> get selectedIndicesBinding {
    _selectedItems =
        (config.viewArguments?.args['selectedItems'] is CVUValueArray)
            ? (config.viewArguments?.args['selectedItems'] as CVUValueArray)
                .value
                .compactMap((CVUValue e) => (e as CVUValueItem).value)
            : [];
    return Binding(
        () => Set.of(selectedItems
            .map((uid) => items.indexWhere((item) => item.id == uid))
            .whereType<int>()
            .toList()), (Set<int> newValue) {
      selectedItems = Set.of(
              newValue.toList().compactMap((index) => items.asMap()[index]?.id))
          .toList();
    });
  }

  // MARK: OTHER
  bool isFullScreen = false;

  // MARK: Configuration
  /// Determines if the query is actively observing for changes. If set to false, it won't respond to changes.
  /// Use this to avoid unneccessary observation if a query isn't being shown in the UI currently
  bool _isObservingQuery = true;

  bool get isObservingQuery => _isObservingQuery;

  set isObservingQuery(bool isObservingQuery) {
    if (_isObservingQuery == isObservingQuery) {
      return;
    }
    _isObservingQuery = isObservingQuery;
    setupQueryObservation();
  }

  /// Sets up a database observation so that all
  setupQueryObservation() {
    var queryConfig = config.query;

    /// Remove old observation
    queryObservation?.cancel();
    queryConfig.removeListener(setupQueryObservation);

    /// Only set up an observation if `isObserving` is true
    if (!isObservingQuery) {
      return;
    }

    queryConfig.addListener(setupQueryObservation);
    if (queryConfig.queryGraphQL != null) {
      getItems(queryConfig);
      //TODO replace it when pod would have sockets
      queryObservation = Stream.periodic(
              Duration(seconds: app.settings.syncControllerIntervalSecs))
          .listen((_) => getItems(queryConfig));
    }
  }

  getItems(DatabaseQueryConfig queryConfig) {
    queryConfig.executeGraphQLRequest().then((records) {
      items = records;
    });
  }

  updateUI() {
    viewDefinition = null;
    _updateCachedValues();
  }
}
