import 'dart:math';

import 'package:flutter/material.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/cvu/constants/cvu_color.dart';
import 'package:memri/core/cvu/models/cvu_value.dart';
import 'package:memri/core/cvu/models/cvu_value_constant.dart';
import 'package:memri/core/cvu/models/cvu_view_arguments.dart';
import 'package:memri/utilities/extensions/collection.dart';
import 'package:memri/widgets/components/shapes/circle.dart';
import 'package:memri/widgets/empty.dart';
import 'package:memri/core/cvu/widgets/renderers/grid_renderer_flow.dart';
import 'package:memri/core/cvu/widgets/renderers/grid_renderer_simple.dart';
import 'package:memri/core/cvu/widgets/renderers/renderer.dart';

/// The grid renderer
/// This presents the data in a grid (aka collection view)
class GridRendererView extends Renderer {
  const GridRendererView({Key? key, required viewContext})
      : super(key: key, viewContext: viewContext);

  @override
  GridRendererViewState createState() => GridRendererViewState();
}

class GridRendererViewState extends RendererViewState {
  late EdgeInsets insets;
  late Point spacing;

  late String layout;

  @override
  initState() {
    super.initState();
    init();

    viewContext.addListener(updateState);
  }

  @override
  void didUpdateWidget(covariant Renderer oldWidget) {
    super.didUpdateWidget(oldWidget);
    setState(() => init());
  }

  @override
  dispose() {
    super.dispose();
    viewContext.removeListener(updateState);
  }

  late Axis scrollDirection;
  late Color backgroundColor;

  updateState() {
    setState(() => init());
  }

  @override
  void init() {
    super.init();
    layout = widget.viewContext.rendererDefinitionPropertyResolver
            .string('layout') ??
        'simple';
    singleChoice =
        viewContext.viewDefinitionPropertyResolver.boolean('singleChoice') ??
            false;
    insets = viewContext.rendererDefinitionPropertyResolver.edgeInsets ??
        EdgeInsets.fromLTRB(5, 80, 5, 5);

    var direction = widget.viewContext.rendererDefinitionPropertyResolver
        .string('scrollDirection');
    scrollDirection =
        direction == 'horizontal' ? Axis.horizontal : Axis.vertical;

    spacing =
        viewContext.rendererDefinitionPropertyResolver.spacing ?? Point(10, 10);
    backgroundColor =
        viewContext.rendererDefinitionPropertyResolver.backgroundColor ??
            CVUColor.system('systemBackground');
  }

  @override
  Widget build(BuildContext context) {
    return ValueListenableBuilder(
        valueListenable: viewContext.itemsValueNotifier,
        builder: (BuildContext context, List<Item> value, Widget? child) {
          if (!viewContext.isLoaded) {
            return Empty();
          }
          List<Widget> elements = [];

          if (viewContext.hasItems) {
            elements = viewContext.items.mapIndexed((index, item) {
              var isSelected = selectedIndices.contains(index);
              return GestureDetector(
                onTap: selectionMode(index),
                child: Stack(
                  alignment: Alignment.topLeft,
                  children: [
                    MouseRegion(
                        cursor: SystemMouseCursors.click,
                        child: viewContext.render(
                            item: item,
                            viewArguments: CVUViewArguments(args: {
                              'isSelected':
                                  CVUValueConstant(CVUConstantBool(isSelected))
                            }))),
                    if (isSelected && showDefaultSelections)
                      SizedBox(
                        height: 30,
                        width: 30,
                        child: Stack(
                          alignment: Alignment.center,
                          children: [
                            Circle(
                                color: Colors.blue,
                                border:
                                    Border.all(color: Colors.white, width: 2)),
                            Icon(
                              Icons.check,
                              color: Colors.white,
                              size: 15,
                            )
                          ],
                        ),
                      )
                  ],
                ),
              );
            }).toList();
          }
          if (startingElement != null) {
            elements.insert(
                0,
                GestureDetector(
                  child: startingElement!,
                ));
          }
          if (trailingElement != null) {
            elements.add(GestureDetector(child: trailingElement!));
          }

          return elements.isNotEmpty
              ? layout == 'flow'
                  ? GridRendererFlowView(elements: elements, spacing: spacing)
                  : GridRendererSimpleView(
                      elements: elements,
                      spacing: spacing,
                      insets: insets,
                      scrollDirection: scrollDirection,
                    )
              : emptyResult ??
                  Padding(
                    padding: EdgeInsets.all(30),
                    child: Center(
                      child: Text(
                        'No items',
                        style: TextStyle(
                            fontSize: 16,
                            fontWeight: FontWeight.normal,
                            color: Color.fromRGBO(0, 0, 0, 0.7),
                            backgroundColor: backgroundColor),
                      ),
                    ),
                  );
        });
  }
}

abstract class GridRenderer extends StatelessWidget {
  final Point spacing;
  final List<Widget> elements;

  const GridRenderer({Key? key, required this.spacing, required this.elements})
      : super(key: key);
}
