import 'package:collection/collection.dart';
import 'package:uuid/uuid.dart';

class Item {
  String type;
  Map<String, dynamic> properties;
  Map<String, EdgeList> edges;

  Item({
    required this.type,
    Map<String, dynamic>? properties,
    Map<String, EdgeList>? edges,
  })  : properties = properties ?? {},
        edges = edges ?? {};

  List<Edge>? getEdges(String edgeName) {
    var edgeList = edges[edgeName];
    if (edgeList == null) {
      return null;
    } else {
      List<Edge> edges = [];
      for (var target in edgeList.targets) {
        edges.add(Edge(
          source: this,
          target: target,
          name: edgeList.name,
        ));
      }
      return edges;
    }
  }

  Item? edgeItem(String edgeName) {
    return edges[edgeName]?.first();
  }

  List<Item>? getEdgeTargets(String edgeName) {
    return edges[edgeName]?.targets;
  }

  dynamic get(String propertyName) => properties[propertyName];

  void setIdIfNotExists() {
    if (get('id') == null) {
      properties['id'] = Uuid().v4();
    }
  }

  get id => get('id');

  void initEdge(String edgeName) {
    if (!edges.containsKey(edgeName)) {
      edges[edgeName] = EdgeList(name: edgeName, targets: []);
    }
  }

  Edge addEdge(String edgeName, Item item) {
    initEdge(edgeName);
    edges[edgeName]!.targets.add(item);
    return Edge(source: this, target: item, name: edgeName);
  }

  factory Item.fromJson(Map<String, dynamic> itemMap) {
    String? type;
    Map<String, dynamic> properties = {};
    Map<String, EdgeList> edges = {};
    itemMap.forEach((key, value) {
      if (key == 'type') {
        type = value;
      } else if (value is List) {
        List<Item> targets = [];
        for (var edgeMap in value) {
          targets.add(Item.fromJson(edgeMap));
        }
        edges[key] = EdgeList(name: key, targets: targets);
      } else {
        properties[key] = value;
      }
    });
    if (type == null) {
      throw Exception('Attempted to create item without an item type.');
    }
    return Item(
      type: type!,
      properties: properties,
      edges: edges,
    );
  }

  Map<String, dynamic> toJson() {
    Map<String, dynamic> result = {};
    result.addAll(properties);
    result['type'] = type;
    return result;
  }
}

class EdgeList {
  // EdgeList is a separate list to support future edge functionalities
  String name;
  List<Item> targets;

  EdgeList({
    required this.name,
    List<Item>? targets,
  }) : targets = targets ?? [];

  Item? first() {
    return targets.firstOrNull;
  }
}

class Edge {
  Item source;
  Item target;
  String name;

  Edge({
    required this.source,
    required this.target,
    required this.name,
  });

  Map<String, dynamic> toJson() {
    return {
      '_source': source.get('id'),
      '_target': target.get('id'),
      '_name': name,
    };
  }
}
