import 'package:memri/constants/app_logger.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:mixpanel_flutter/mixpanel_flutter.dart';

class AnalyticsService {
  static Mixpanel? mixpanel;

  Future<void> init() async {
    try {
      mixpanel ??= await Mixpanel.init(
        app.settings.mixPanelToken,
        optOutTrackingDefault: false,
      );
    } catch (e) {
      AppLogger.err(
        'Unable to connect to Mixpanel',
        error: e,
        sendCrashReport: false,
      );
    }
  }

  void reset() => mixpanel!.reset();

  void logSignUp(String userId, bool developersMode) {
    mixpanel!.timeEvent(AnalyticsEvents.signUp);
    mixpanel!.identify(userId);
    mixpanel!.track(AnalyticsEvents.signUp,
        properties: {_AnalyticsProperties.developersMode: developersMode});
  }

  void logSignIn(String userId, bool developersMode) {
    mixpanel!.timeEvent(AnalyticsEvents.signIn);
    mixpanel!.identify(userId);
    mixpanel!.track(AnalyticsEvents.signIn,
        properties: {_AnalyticsProperties.developersMode: developersMode});
  }

  void logRouteNavigation(String route) {
    if (mixpanel != null) {
      mixpanel!.timeEvent(AnalyticsEvents.routeNavigation);
      mixpanel!.track(AnalyticsEvents.routeNavigation,
          properties: {_AnalyticsProperties.screen: route});
    }
  }

  void logTobBarTabButton(String buttonText) {
    mixpanel!.timeEvent(AnalyticsEvents.breadCrumbButton);
    mixpanel!.track(AnalyticsEvents.breadCrumbButton,
        properties: {_AnalyticsProperties.buttonLabel: buttonText});
  }

  void logCvuButton(String buttonText) {
    mixpanel!.timeEvent(AnalyticsEvents.cvuButton);
    mixpanel!.track(AnalyticsEvents.cvuButton,
        properties: {_AnalyticsProperties.buttonLabel: buttonText});
  }

  void logNavigationButton(String target) {
    mixpanel!.timeEvent(target);
    mixpanel!.track(target);
  }

  void logImporterSelect(String importerName) {
    mixpanel!.timeEvent(AnalyticsEvents.importerSelect);
    mixpanel!.track(AnalyticsEvents.importerSelect,
        properties: {_AnalyticsProperties.importerName: importerName});
  }

  void logImporterConnect(String importerName) {
    mixpanel!.timeEvent(AnalyticsEvents.importerConnect);
    mixpanel!.track(AnalyticsEvents.importerConnect,
        properties: {_AnalyticsProperties.importerName: importerName});
  }

  void logImporterStatus(String importerStatus) {
    mixpanel!.timeEvent(AnalyticsEvents.importerStatus);
    mixpanel!.track(AnalyticsEvents.importerStatus,
        properties: {_AnalyticsProperties.importerStatus: importerStatus});
  }

  void logProjectCreate(String projectName, var projectDataSource) {
    mixpanel!.timeEvent(AnalyticsEvents.projectCreate);
    mixpanel!.track(
      AnalyticsEvents.projectCreate,
      properties: {
        _AnalyticsProperties.projectName: projectName,
        _AnalyticsProperties.projectDataSource: projectDataSource,
      },
    );
  }

  void logProjectDataSelect(var dataSources) {
    mixpanel!.timeEvent(AnalyticsEvents.projectDataSelect);
    mixpanel!.track(
      AnalyticsEvents.projectDataSelect,
      properties: {
        _AnalyticsProperties.projectDataSource: dataSources.toString()
      },
    );
  }

  void logProjectAddLabels(List<String> labels) {
    mixpanel!.timeEvent(AnalyticsEvents.projectLabelsAdd);
    mixpanel!.track(
      AnalyticsEvents.projectLabelsAdd,
      properties: {_AnalyticsProperties.projectDataLabels: labels},
    );
  }

  void logProjectLabelsOverview(
      String submitted, String skipped, String total) {
    mixpanel!.timeEvent(AnalyticsEvents.projectLabelsOverview);
    mixpanel!.track(
      AnalyticsEvents.projectLabelsOverview,
      properties: {
        _AnalyticsProperties.projectLabelsSubmitted: submitted,
        _AnalyticsProperties.projectLabelsSkipped: skipped,
        _AnalyticsProperties.projectLabelsTotal: total,
      },
    );
  }

  void logProjectLabelsSummary(List<String> labels) {
    mixpanel!.timeEvent(AnalyticsEvents.projectLabelsSummary);
    mixpanel!.track(
      AnalyticsEvents.projectLabelsSummary,
      properties: {_AnalyticsProperties.projectLabelsTotal: labels},
    );
  }

  void logProjectTrainModel() {
    mixpanel!.timeEvent(AnalyticsEvents.projectTrainModel);
    mixpanel!.track(AnalyticsEvents.projectTrainModel);
  }

  void logProjectGoogleColab() {
    mixpanel!.timeEvent(AnalyticsEvents.projectGoogleColab);
    mixpanel!.track(AnalyticsEvents.projectGoogleColab);
  }

  void logProjectPluginGitlabUrl(String url) {
    mixpanel!.timeEvent(AnalyticsEvents.projectPluginGitlabUrl);
    mixpanel!.track(
      AnalyticsEvents.projectPluginGitlabUrl,
      properties: {_AnalyticsProperties.projectPluginGitlabUrl: url},
    );
  }

  void logProjectTutorialLink(String linkLabel, String linkUrl) {
    mixpanel!.timeEvent(AnalyticsEvents.projectTutorialLink);
    mixpanel!.track(
      AnalyticsEvents.projectTutorialLink,
      properties: {
        _AnalyticsProperties.projectTutorialLinkLabel: linkLabel,
        _AnalyticsProperties.projectTutorialLinkUrl: linkUrl,
      },
    );
  }

  Future<void> logError(String errorName, String stackTrace) async {
    await init();
    if (mixpanel != null) {
      mixpanel!.track(
        AnalyticsEvents.error,
        properties: {
          _AnalyticsProperties.errorName: errorName,
          _AnalyticsProperties.errorStackTrace: stackTrace,
        },
      );
    }
  }
}

class AnalyticsEvents {
  AnalyticsEvents._();

  static const signUp = 'sign_up';
  static const signIn = 'sign_in';
  static const signOut = 'sign_out';
  static const error = 'error';

  static const routeNavigation = 'route_navigation';
  static const breadCrumbButton = 'bread_crumb_button';
  static const cvuButton = 'cvu_button';

  static const discord = 'discord';
  static const discordPlugin = 'discord_plugin';
  static const gitlab = 'gitlab';
  static const guide = 'guide';
  static const documentation = 'documentation';
  static const privacy = 'privacy_policy';
  static const contact = 'contact_us';
  static const dataExplorer = 'data_explorer';
  static const pluginDoc = 'documentation_plugin';
  static const customizeApp = 'customize_app';

  static const importerSelect = 'importer_select';
  static const importerConnect = 'importer_connect';
  static const importerStatus = 'importer_status';

  static const projectCreate = 'project_create';
  static const projectDataSelect = 'project_data_select';
  static const projectLabelsAdd = 'project_labels_add';
  static const projectLabelsOverview = 'project_labels_overview';
  static const projectLabelsSummary = 'project_labels_summary';
  static const projectTrainModel = 'project_train_model';
  static const projectDeployApp = 'project_deploy_app';
  static const projectGoogleColab = 'project_google_colab';
  static const projectTutorialLink = 'project_tutorial_link';
  static const projectPluginGitlabUrl = 'project_plugin_gitlab_url';
}

class _AnalyticsProperties {
  _AnalyticsProperties._();

  static const buttonLabel = 'button_label';
  static const developersMode = 'developers_mode';
  static const screen = 'screen';

  static const importerName = 'importer_name';
  static const importerStatus = 'importer_status';

  static const projectName = 'project_name';
  static const projectDataSource = 'project_date_source';
  static const projectDataLabels = 'project_date_labels';
  static const projectLabelsSubmitted = 'project_labels_submitted';
  static const projectLabelsSkipped = 'project_labels_skipped';
  static const projectLabelsTotal = 'project_labels_total';
  static const projectPluginGitlabUrl = 'project_plugin_gitlab_url';
  static const projectTutorialLinkLabel = 'project_tutorial_link_label';
  static const projectTutorialLinkUrl = 'project_tutorial_link_url';

  static const errorName = 'error_name';
  static const errorStackTrace = 'error_stack_trace';
}
