import 'dart:convert';
import 'dart:math';

import 'package:flutter/services.dart';
import 'package:get_it/get_it.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/controllers/app_controller.dart';
import 'package:memri/core/controllers/database_controller.dart';
import 'package:memri/core/controllers/file_storage/file_storage_controller.dart';
import 'package:memri/core/models/database/item_edge_record.dart';
import 'package:memri/core/models/database/item_property_record.dart';
import 'package:memri/core/models/database/item_record.dart';
import 'package:memri/core/services/database/property_database_value.dart';
import 'package:memri/core/services/database/schema.dart';
import 'package:memri/utilities/extensions/string.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:uuid/uuid.dart';

class DemoData {
  static Map<String, SchemaType> types = {};

  static importSchemaOnce(
      {DatabaseController? databaseController,
      bool throwIfAgainstSchema = false}) async {
    databaseController ??= AppController.shared.databaseController;

    var schemaItems = await ItemRecord.fetchWithType(
        'ItemPropertySchema', databaseController.databasePool);
    if (schemaItems.isEmpty) {
      schemaItems = await ItemRecord.fetchWithType(
          'ItemEdgeSchema', databaseController.databasePool);
    }

    if (schemaItems.isNotEmpty) {
      return;
    }

    var fileURL = 'assets/schema.json';
    var fileData = await rootBundle.loadString(fileURL, cache: false);
    var items = jsonDecode(fileData);
    if (items == null ||
        items['properties'] == null ||
        items['properties'] is! List ||
        items['edges'] == null ||
        items['edges'] is! List) {
      throw Exception('Could not locate schema file');
    }

    await databaseController.databasePool.schemaImportTransaction(items);
    await GetIt.I<Schema>().load();
  }

  static importDefaultData(
      {DatabaseController? databaseController,
      bool throwIfAgainstSchema = false}) async {
    databaseController ??= AppController.shared.databaseController;
    await importData(
        fileName: app.settings.defaultDatabase,
        databaseController: databaseController,
        throwIfAgainstSchema: throwIfAgainstSchema);
  }

  static importDemoData(
      {DatabaseController? databaseController,
      bool throwIfAgainstSchema = false}) async {
    databaseController ??= AppController.shared.databaseController;
    await importData(
        fileName: 'demo_database',
        databaseController: databaseController,
        throwIfAgainstSchema: throwIfAgainstSchema);
  }

  static Future<void> importData(
      {required String fileName,
      required DatabaseController databaseController,
      bool throwIfAgainstSchema = false}) async {
    var fileURL = 'assets/$fileName.json';
    var fileData = await rootBundle.loadString(fileURL, cache: false);
    var items = jsonDecode(fileData);
    if (items == null || items is! List) {
      return;
    }
    await databaseController.databasePool.transaction(() async {
      List<DemoDataItem> processedItems = (await Future.wait(items.map(
              (item) async => await processItemJSON(
                  item: item, isRunningTests: throwIfAgainstSchema))))
          .expand((element) => element)
          .toList();

      Map<String, String> tempIDLookup = {};
      Map<String, String> sourceIDLookup = {};
      List<ItemRecord> records = [];
      List<ItemPropertyRecord> properties = [];
      List<ItemEdgeRecord> edges = [];

      for (var item in processedItems) {
        var record = ItemRecord(
            uid: item.uid,
            type: item.type,
            dateCreated: item.dateCreated,
            dateModified: item.dateModified);
        var tempUID = item.tempUID;

        if (record.type == 'File') {
          record.fileState = FileState.needsUpload;
        }
        records.add(record);

        if (tempUID != null) {
          tempIDLookup[tempUID] = record.uid;
        }
        sourceIDLookup[item.uid] = record.uid;
      }

      for (var item in processedItems) {
        for (var property in item.properties) {
          ItemPropertyRecord record = ItemPropertyRecord(
              itemUID: item.uid, name: property.name, value: property.value);
          properties.add(record);
        }
        for (var edge in item.edges) {
          var targetActualID = tempIDLookup[edge.targetTempUID];
          if (targetActualID == null) {
            continue;
          }

          var sourceUID = sourceIDLookup[item.uid];

          var record = ItemEdgeRecord(
              sourceUID: sourceUID, name: edge.name, targetUID: targetActualID);
          edges.add(record);
        }
      }
      await ItemRecord.insertList(records, db: databaseController.databasePool);
      await ItemPropertyRecord.insertList(properties,
          db: databaseController.databasePool);
      await ItemEdgeRecord.insertList(edges,
          db: databaseController.databasePool);
    });
  }

  static Future<List<DemoDataItem>> processItemJSON(
      {required Map<String, dynamic> item,
      String? overrideUID,
      isRunningTests = false}) async {
    handleError(String string) {
      if (isRunningTests) {
        // Used for testing: throw an error if error in demo data
        throw (string);
      } else {
        // Notify developer of error in demo data, but continue
        AppLogger.err(string);
      }
    }

    var itemType = item['_type'];
    if (itemType is! String) {
      handleError('BAD RECORD: $item');
      return [];
    }

    if (types[itemType] == null) {
      AppLogger.warn('$itemType not in schema');
      return [];
    }

    var itemTempUID = overrideUID ?? item['uid']?.toString();

    List<DemoDataItem> items = [];
    List<DemoDataProperty> properties = [];
    List<DemoDataEdge> edges = [];
    // Fake a recent date for the demo data

    var dateCreated = DateTime.now()
        .subtract(Duration(milliseconds: Random().nextInt(1814400 * 1000)));
    var dateModified = dateCreated;

    await Future.forEach(item.entries, (MapEntry itemProperty) async {
      var propertyName = itemProperty.key;
      var propertyValue = itemProperty.value;
      switch (propertyName) {
        case '_type':
        case 'uid':
        case 'version':
          return;
        case 'allEdges':
          var edgeList = propertyValue;
          if (edgeList is! List ||
              edgeList.isEmpty ||
              (edgeList.isNotEmpty && edgeList[0] is! Map)) {
            return;
          }
          for (var edge in edgeList) {
            var edgeName = edge['_type'];
            if (edgeName is! String) {
              return;
            }

            SchemaEdge? expectedEdge = types[itemType]?.edgeTypes[edgeName];
            if (expectedEdge == null) {
              handleError('$itemType.$edgeName edge not in schema');
              return;
            }
            String? targetType = edge['targetType'] ?? edge['_target']['_type'];
            if (targetType == null) {
              handleError('$itemType.$edgeName targetType is missing');
              return;
            }
            if (targetType != expectedEdge.targetType &&
                expectedEdge.targetType != 'Any') {
              handleError(
                  '$itemType.$edgeName targetType should be ${expectedEdge.targetType}, $targetType received');
              return;
            }

            var targetUID = edge['uid']?.toString();
            if (targetUID != null) {
              edges.add(DemoDataEdge(
                  name: edgeName,
                  targetTempUID: targetUID,
                  targetType: targetType));
            } else {
              var subitem = edge['_target'];
              if (subitem is Map) {
                // Sub-item declared as edge, add edge property AND item
                var targetUID = Uuid().v4();
                edges.add(DemoDataEdge(
                    name: edgeName,
                    targetTempUID: targetUID,
                    targetType: targetType));
                items = [
                  ...items,
                  ...await processItemJSON(
                      item: subitem as Map<String, dynamic>,
                      overrideUID: targetUID,
                      isRunningTests: isRunningTests)
                ];
              }
            }
          }
          break;
        case 'dateCreated':
          var dateInMilliseconds = propertyValue;
          if (dateInMilliseconds is! int) return;
          var date = DateTime.fromMillisecondsSinceEpoch(dateInMilliseconds);
          dateCreated = date;
          dateModified = date;
          break;
        default:
          var expectedType = types[itemType]?.propertyTypes[propertyName];
          if (expectedType == null) {
            handleError('$itemType.$propertyName property not in schema');
            return;
          }
          try {
            var databaseValue = PropertyDatabaseValue.create(propertyValue,
                expectedType.valueType, '$itemType.$propertyName');

            if (!isRunningTests &&
                itemType == 'File' &&
                propertyName == 'filename' &&
                databaseValue is PropertyDatabaseValueString) {
              var fileName = databaseValue.value;

              var demoDirectory = 'assets/demoAssets';
              var sourcePath =
                  '$demoDirectory/${fileName.fileName ?? ""}.${fileName.fileExtension ?? "jpg"}';

              // Also add sha256 property for item
              var byteData =
                  await FileStorageController.getByteDataFromAsset(sourcePath);
              var sha256 = FileStorageController.getHashForData(byteData.buffer
                  .asUint8List(byteData.offsetInBytes, byteData.lengthInBytes));

              var url = (await FileStorageController.getFileStorageURL()) +
                  '/' +
                  sha256;

              properties.add(DemoDataProperty(
                  name: 'sha256', value: PropertyDatabaseValueString(sha256)));

              await FileStorageController.copy(sourcePath, url);
            }

            properties.add(
                DemoDataProperty(name: propertyName, value: databaseValue));
          } catch (error) {
            handleError(error.toString());
            return;
          }
      }
    });

    items.add(DemoDataItem(
        type: itemType,
        tempUID: itemTempUID,
        properties: properties,
        edges: edges,
        dateCreated: dateCreated,
        dateModified: dateModified));
    return items;
  }
}

class DemoDataItem {
  String type;
  String uid;
  String? tempUID;
  List<DemoDataProperty> properties;
  List<DemoDataEdge> edges;
  DateTime? dateCreated;
  DateTime? dateModified;

  DemoDataItem({
    required this.type,
    uid,
    this.tempUID,
    required this.properties,
    required this.edges,
    this.dateCreated,
    this.dateModified,
  }) : uid = uid ?? Uuid().v4().toString();
}

class DemoDataProperty {
  String name;
  PropertyDatabaseValue value;

  DemoDataProperty({required this.name, required this.value});
}

class DemoDataEdge {
  String name;
  String targetTempUID;
  String targetType;

  DemoDataEdge(
      {required this.name,
      required this.targetTempUID,
      required this.targetType});
}

class SchemaFile {
  List<SchemaProperty> properties;
  List<SchemaEdge> edges;

  SchemaFile(this.properties, this.edges);

  SchemaFile.fromJson(Map<String, dynamic> json)
      : properties = (json['properties'] as List)
            .map((property) => SchemaProperty.fromJson(property))
            .toList(),
        edges = (json['edges'] as List)
            .map((edge) => SchemaEdge.fromJson(edge))
            .toList();

  Map<String, dynamic> toJson() => {
        'properties': properties,
        'edges': edges,
      };
}
