import 'package:memri/core/apis/pod_api.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/models/pod/pod_config.dart';
import 'package:memri/core/services/api_service.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/network_helper.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:collection/collection.dart';

class GraphService extends ApiService<PodAPI> {
  GraphService(this._prefs) : super(api: PodAPI()) {
    _setConfig();
  }

  final SharedPreferences _prefs;
  late PodConfig podConfig;

  Future<List<Item>> getResultAsItemList({required String query}) async {
    _setConfig();
    return NetworkHelper().parseGQLResponse(await api.queryGraphQL(query));
  }

  void _setConfig() {
    podConfig = PodConfig(
      baseUrl:
          _prefs.getString(app.keys.podAddress) ?? app.settings.defaultPodUrl,
      ownerKey: _prefs.getString(app.keys.ownerKey) ?? '',
      databaseKey: _prefs.getString(app.keys.dbKey) ?? '',
      apiVersion: app.settings.podVersion,
    );
    api.setConnectionConfig(podConfig);
  }

  Future<List<Item>> fetchActiveImporters() async =>
      await getResultAsItemList(query: '''
query {
        PluginRun {
          id
          pluginName
          status
        }
      }
''');

  Future<List<Item>> checkIfDataInPod(String source) async =>
      await getResultAsItemList(query: '''
query {
        Message(limit: 1, filter: {service: {eq: "$source"}}) {
          id
          service
        }
      }
''');

  Future<List<Item>> fetchWhatsappMessages() async =>
      await getResultAsItemList(query: '''
query {
        Message(limit: 1, filter: {service: {eq: "whatsapp"}}) {
          id
          service
        }
      }
''');

  Future<List<Item>> fetchSampleWhatsappMessages() async =>
      await getResultAsItemList(query: '''
query {
        Message(limit: 1, filter: {service: {eq: "sample-whatsapp"}}) {
          id
          service
        }
      }
''');

  Future<List<Item>> fetchTwitterTweets() async =>
      await getResultAsItemList(query: '''
query {
        Tweet(limit: 1) {
          id
        }
      }
''');

  Future<List<Item>> getOAuthFlowItems(String service) async =>
      await getResultAsItemList(query: '''
      query {
        OauthFlow (limit: 500 filter: {service: {eq: "$service"}}) {
          accessToken
          refreshToken
          dateCreated
          service
        }
      }''');

  Future<List<Item>> getProjectItems(String projectId) async =>
      await getResultAsItemList(query: '''
      query {
        Project (filter: {id: {eq: "$projectId"}}) {
          name
          dataset {
            name
            entry {
              skippedByLabeller
              data {
                content
                message
                dateSent
              }
              annotation {
                labelValue
              }
            }

            labellingTask {
              id
              labelOption {
                name
              }
            }
          }
        }
      }
    ''');

  Future<List<Item>> getAllProjectsNames() async =>
      await getResultAsItemList(query: '''
      query {
        Project (filter: {deleted: {eq: 0}}){
          name
        }
      }''');

  Future<List<Item>> getProjectSummaries() async =>
      await getResultAsItemList(query: '''
      query {
        Project(limit: 1000, filter: {deleted: {eq: 0}}) {
          id
          name
          dataset (limit: 1) {
            name
            entry {
              skippedByLabeller
              annotation {
                labelValue
              }
            }
          }
        }
      }''');

  Future<List<Item>> getMessages(String service) async =>
      await getResultAsItemList(query: '''
      query {
        Message (limit: 500, filter: {service: {eq: "$service"}}) {
          id
          content
          dateSent
          service
        }
      }''');

  Future<List<Item>> getTweets() async => await getResultAsItemList(query: '''
      query {
        Tweet (limit: 500) {
          id
          message
          postDate
        }
      }''');

  Future<List<Item>> getMockMessages(String sourceProject) async =>
      await getResultAsItemList(query: '''
      query {
        Message(order_desc: dateSent, filter: {sourceProject: {eq: "$sourceProject"}}) {
          content
          dateSent
          sourceProject
        }
      }
    ''');

  Future<List<Item>> loadMessages(int amount) async =>
      await getResultAsItemList(query: '''
      query {
        Message(limit: $amount, order_desc: dateSent) {
          dateSent
          content
        }
      }
    ''');

  Future<List<Item>> loadAppContent(String pluginId) async =>
      await getResultAsItemList(query: '''
      query {
        Message(order_desc: dateSent, filter: {sourceProject: {eq: "$pluginId"}}) {
          dateSent
          content
          sourceProject
          label {
            value
          }
        }
      }
    ''');

  Future<List<Item>> fetchDataApps() async =>
      await getResultAsItemList(query: '''
      query {
        Plugin {
          containerImage
          configJson
          config
          dataType
          pluginDescription
          name
          pluginModule
          pluginName
          pluginType
          gitProjectId
        }
      }
    ''');

  Future<List<Item>> getInboxChats() async =>
      await getResultAsItemList(query: '''
      query {
        MessageChannel (limit: 1000, filter: {service: {eq: "whatsapp"}}) {
          id
          name
          service
          photo {
            id
            file {
              sha256
            }
          }
          ~messageChannel (limit: 1, order_desc: dateSent) {
            dateSent
          }
        }
      }''');

  Future<List<Item>> getInboxMessages(
          String chatID, int limit, int offset) async =>
      await getResultAsItemList(query: '''
      query {
        MessageChannel (filter: {id: {eq: "$chatID"}}) {
          id
          ~messageChannel (limit: $limit, offset: $offset, order_desc: dateSent) {
            dateSent
            content
            sender {
              displayName
                isMe
                profilePicture {
                id
                file {
                  sha256
                }
              }
            }
          }
        }
      }
    ''');

  Future<List<Item>> getTwitterFeed() async =>
      await getResultAsItemList(query: '''
      query {
        Tweet (limit: 1000, order_desc: postDate) {
          externalId
          message
          author {
            handle
            displayName
            profilePicture {
                id
                file {
                  sha256
                }
            }
          }
          mention {
            handle
            displayName
          }
          postDate
        }
      }''');

  Future<Item?> getPropertySchema(String itemType, String propertyName) async {
    List<Item> result = await getResultAsItemList(query: '''
      query {
        ItemPropertySchema(limit: 1, filter: {and: [
              {itemType: {eq: "$itemType"}},
              {propertyName: {eq: "$propertyName"}}
            ]}) {
          itemType
          propertyName
        }
      }
      ''');
    return result.firstOrNull;
  }
}
