import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/models/running_plugin.dart';
import 'package:memri/core/services/gitlab_service.dart';
import 'package:memri/core/services/graph_service.dart';
import 'package:memri/core/services/pod_service.dart';
import 'package:memri/utilities/helpers/time_helper.dart';
import 'package:visibility_detector/visibility_detector.dart';

enum AppState { loading, showApp, showLogs, showSettings }

extension ParseToString on AppState {
  String toUrlString() {
    switch (this) {
      case AppState.loading:
        return 'loading';
      case AppState.showApp:
        return 'app';
      case AppState.showLogs:
        return 'logs';
      case AppState.showSettings:
        return 'settings';
    }
  }
}

enum AppListState { fetching, fetched }

class DataAppProvider with ChangeNotifier {
  final PodService _podService;
  final GraphService _graphService;
  final GitlabService _gitlabService;

  DataAppProvider(this._podService, this._graphService, this._gitlabService);

  // TODO handle schema checking somewhere else
  final List<String> _runProperties = [
    'containerImage',
    'authUrl',
    'pluginModule',
    'pluginName',
    'status',
    'targetItemId',
    'error',
    'config',
    'progress',
  ];

  final List<String> _pluginProperties = [
    'bundleImage',
    'containerImage',
    'configJson',
    'config',
    'dataType',
    'icon',
    'pluginDescription',
    'name',
    'pluginModule',
    'pluginName',
    'pluginType',
    'gitProjectId',
  ];

  // ignore: cancel_subscriptions
  StreamSubscription<List<Item>>? feedStreamSubscription;
  List<Item> feed = [];

  String getFeedPostDate(int feedIndex) =>
      DateTimeHelper().getDateString(feed[feedIndex].get('postDate'));

  String getFeedAuthorName(int feedIndex) =>
      feed[feedIndex].edgeItem('author')?.get('displayName') ?? '';

  String getFeedAuthorHandle(int feedIndex) =>
      feed[feedIndex].edgeItem('author')?.get('handle') ?? '';

  reset() {
    currentPluginRun = null;
    currentPluginRunStream = null;
    currentAppState = AppState.loading;
    currentRunLog = null;
    currentGitProjectId = null;
    currentPluginMessages = [];
    runningPlugins = {};

    //Inbox
    chatStream = null;
    messageStream = null;
    if (chatStreamSubscription != null) chatStreamSubscription!.cancel();
    if (messageStreamSubscription != null) messageStreamSubscription!.cancel();
    selectedChannel = -1;
    chats = [];
    activeChatMessages = [];
    _loadingMessages = false;
    chatImageSha = {};
    chatImageProviders = {};
    messageImageSha = {};
    messageImageProviders = {};
    selectedChatRoomID = null;
  }

  /*
   * Inbox
   */

  Future<Uint8List> downloadPhoto(String fileSHAHash) async {
    return await _podService.downloadFile(fileSHAHash);
  }

  /*
  * Running a plugin
  */

  Item? currentPluginRun;
  Stream<Item>? currentPluginRunStream;
  StreamSubscription<Item>? currentPluginRunSubscription;
  AppState currentAppState = AppState.loading;
  AppListState appListState = AppListState.fetching;
  String? currentRunLog;
  int? currentGitProjectId;
  List<Item> currentPluginMessages = [];

  setDataAppState(AppState state) {
    currentAppState = state;

    /// TODO: FIX THIS, CAUSE EXCEPTION
    notifyListeners();
  }

  handlePublishApp(BuildContext context, String projectName) {
    // TODO delete previous app with name + gitProjectId
    createPluginItem(projectName);
    RouteNavigator.navigateTo(context: context, route: Routes.apps);
  }

  handleRerunApp() {
    if (currentGitProjectId != null) {
      currentPluginRun = null;
      notifyListeners();
      runPluginFromGit(gitProjectId: currentGitProjectId!, changeState: false);
    }
  }

  createPluginItem(String projectName) async {
    // Create data app from current gitlab id
    var metadata = await getMetadata(currentGitProjectId!);
    var plugin = Item(type: 'Plugin');
    metadata.forEach((key, value) {
      if (_pluginProperties.contains(key)) {
        if (key == 'containerImage' && value is String) {
          value = value.toLowerCase();
        }
        plugin.properties[key] = value;
      }
    });

    plugin.properties['name'] = projectName;
    plugin.properties['pluginDescription'] = null;
    plugin.properties['gitProjectId'] = currentGitProjectId;
    plugin.properties['dataType'] = 'Text';
    plugin.properties['pluginType'] = 'dataApp';

    await _podService.createItem(item: plugin);
    dataApps.add(plugin);
    AppLogger.debug('Created plugin: $projectName');
  }

  runPluginFromGit({required int gitProjectId, bool changeState = true}) async {
    currentGitProjectId = gitProjectId;
    currentPluginRun = null;
    currentPluginMessages =
        await createMockMessages(10, '_preview-$gitProjectId');

    var pluginRun = await createPluginRunFromGit(gitProjectId);
    await _podService.createItem(item: pluginRun);
    AppLogger.debug('Running plugin: ${pluginRun.toJson()}');

    currentPluginRun = pluginRun;
    setupPluginRunStream(
        runId: currentPluginRun!.get('id'), changeState: changeState);
  }

  Stream<Item> _pluginRunStream(String runId) async* {
    while (true) {
      var run = await _podService.getItem(id: runId);
      if (run != null) {
        yield run;
      }
      await Future.delayed(Duration(seconds: 1));
    }
  }

  bool runIsDone(String? runStatus) => ['done', 'error'].contains(runStatus);

  setupPluginRunStream({required String runId, bool changeState = true}) {
    currentPluginRunStream = _pluginRunStream(runId);
    currentPluginRunSubscription = currentPluginRunStream!.listen((run) async {
      currentPluginRun = run;
      AppLogger.debug('run status: ${getCurrentRunStatus()}');
      notifyListeners();

      if (runIsDone(getCurrentRunStatus())) {
        stopPluginRunStream();
        // loadCurrentAppContent();
        if (changeState) {
          currentAppState = AppState.showApp;
        }
        getCurrentRunLog();
        currentPluginMessages =
            await loadAppContent('_preview-$currentGitProjectId');
        for (var msg in currentPluginMessages) {
          if (msg.edgeItem('label') != null) {
            AppLogger.debug(msg.edgeItem('label')!.get('value'));
          }
        }
        notifyListeners();
        AppLogger.debug('Run done');
        AppLogger.debug('app message: ${currentPluginMessages.length}');
        AppLogger.debug('has log: ${currentRunLog != null}');
      }
    });
  }

  stopPluginRunStream() {
    currentPluginRunSubscription?.cancel();
    currentPluginRunStream = null;
  }

  Future<Map<String, dynamic>> getMetadata(int gitProjectId) async {
    // TODO error handling if metadata does not exist
    AppLogger.debug('fetching metadata...');
    var fileContent = await _gitlabService.getTextFileContent(
        projectId: gitProjectId, filepath: 'metadata.json');
    return jsonDecode(fileContent);
  }

  Future<Item> createPluginRunFromGit(int gitProjectId) async {
    var metadata = await getMetadata(gitProjectId);

    var pluginRun = Item(type: 'PluginRun');
    metadata.forEach((key, value) {
      if (_runProperties.contains(key)) {
        // containerImage needs to be lowercase
        if (key == 'containerImage' && value is String) {
          value = value.toLowerCase();
        }
        pluginRun.properties[key] = value;
      }
    });

    // Set id separately, pluginrun needs targetItemId == id
    pluginRun.setIdIfNotExists();
    pluginRun.properties['targetItemId'] = pluginRun.get('id');

    pluginRun.properties['config'] = jsonEncode({'isMock': true});
    return pluginRun;
  }

  getCurrentRunLog() async {
    if (currentPluginRun != null) {
      currentRunLog = await getRunLog(currentPluginRun!);
    }
  }

  Future<String?> getRunLog(Item pluginRun) async {
    String? runId = pluginRun.get('id');
    if (runId != null) {
      return await _podService.getLogsForPluginRun(pluginRun);
    }
    return null;
  }

  bool pluginIsReady() {
    if (getCurrentRunStatus() == 'done') {
      return true;
    }
    return false;
  }

  String? getCurrentRunStatus() {
    // TODO run status enum, sync with pymemri.plugin.states
    return currentPluginRun?.get('status');
  }

  Future<List<Item>> createMockMessages(
      int amount, String sourceProject) async {
    var existingMessages = await _graphService.getMockMessages(sourceProject);
    Set<String> existingContent = {};
    for (var msg in existingMessages) {
      existingContent.add(msg.get('content'));
    }

    var messages = await _graphService.loadMessages(amount);
    List<Item> newMessages = [];
    for (var message in messages) {
      String? content = message.get('content');
      if (content != null && !existingContent.contains(content)) {
        var newMessage = Item.fromJson({
          'type': 'Message',
          'content': message.get('content'),
          'dateSent': message.get('dateSent'),
          'isMock': true,
          'sourceProject': sourceProject,
        });
        newMessages.add(newMessage);
      }
    }

    await _podService.bulkAction(createItems: newMessages);
    AppLogger.debug('Created ${newMessages.length} new mock messages');
    return newMessages;
  }

  Future<List<Item>> loadAppContent(String pluginId,
      {bool onlyWithPredictions = true}) async {
    var response = await _graphService.loadAppContent(pluginId);

    List<Item> result;
    if (onlyWithPredictions) {
      result = [];
      for (var element in response) {
        if (element.edgeItem('label')?.get('value') != null) {
          result.add(element);
        }
      }
    } else {
      result = response;
    }

    AppLogger.debug('loaded ${result.length} messages');
    notifyListeners();

    return result;
  }

  /*
  * Apps screen
  */

  // pluginId: runningPlugin
  Map<String, RunningPlugin> runningPlugins = {};
  List<Item> dataApps = [];

  fetchDataApps() async {
    _handleAppListFetching();

    dataApps =
        await _graphService.fetchDataApps().whenComplete(_handleAppListFetched);
    getRunningPlugins();
    notifyListeners();
  }

  handleOpenDataApp(BuildContext context, Item plugin, AppState appState) {
    var pluginId = plugin.get('id');
    if (!runningPlugins.containsKey(pluginId)) {
      runPlugin(plugin);
    } else {
      AppLogger.debug('Already running');
    }
    RouteNavigator.navigateTo(
        route: Routes.appConfigure,
        context: context,
        param: {'id': pluginId, 'tab': appState.toUrlString()});
  }

  runPlugin(Item plugin) {
    AppLogger.debug('Creating pluginRun: ${plugin.get("name")}');
    Item pluginRun = Item(type: 'PluginRun');
    plugin.properties.forEach((key, value) {
      if (_runProperties.contains(key)) {
        pluginRun.properties[key] = value;
      }
    });
    pluginRun.setIdIfNotExists();
    pluginRun.properties['targetItemId'] = pluginRun.get('id');
    pluginRun.properties['config'] = jsonEncode({'isMock': true});

    createMockMessages(128, plugin.get('id'));
    _podService.createItem(item: pluginRun);
    AppLogger.debug("created PluginRun ${pluginRun.get("id")}");
    _createRunningPlugin(plugin, pluginRun);
  }

  _createRunningPlugin(Item plugin, Item pluginRun) {
    var runningPlugin = RunningPlugin(plugin: plugin, pluginRun: pluginRun);
    runningPlugins[plugin.get('id')] = runningPlugin;

    runningPlugin.pluginRunStream = _pluginRunStream(pluginRun.get('id'));
    AppLogger.debug(runningPlugin.pluginRunStream.toString());
    runningPlugin.pluginRunSubscription =
        runningPlugin.pluginRunStream!.listen((run) async {
      runningPlugin.pluginRun = run;

      if (runningPlugin.isDone()) {
        AppLogger.debug('Run done');
        runningPlugin.runLog = await getRunLog(runningPlugin.pluginRun);

        // Stop plugin
        runningPlugin.stopStream();
        runningPlugins.remove(runningPlugin.plugin.get('id'));
      }

      notifyListeners();
    });

    AppLogger.debug("created running plugin ${plugin.get("name")}");
  }

  void updateProjectSettings(
      RunningPlugin? runningPlugin, String? name, String? description) {
    if (runningPlugin == null) {
      return;
    }
    var plugin = runningPlugin.plugin;
    if (name != null && name.isNotEmpty) {
      plugin.properties['name'] = name;
    }
    // TODO error for empty name
    plugin.properties['pluginDescription'] = description;
    _podService.updateItem(item: plugin);
    notifyListeners();
  }

  _handleAppListFetching() {
    appListState = AppListState.fetching;
    notifyListeners();
  }

  _handleAppListFetched() {
    appListState = AppListState.fetched;
    notifyListeners();
  }

  setAppState(RunningPlugin? runningPlugin, AppState state) {
    if (runningPlugin != null) {
      runningPlugin.appState = state;
      notifyListeners();
    }
  }

  void getRunningPlugins() {
    // TODO get all running data apps from the pod to initialize apps screen
  }

  updateRunningPlugin(String pluginId) {
    if (dataApps.isNotEmpty) {
      for (Item dataApp in dataApps) {
        if (dataApp.get('id') == pluginId) {
          if (!runningPlugins.containsKey(pluginId)) {
            runPlugin(dataApp);
          }
        }
        break;
      }
    }
  }

  RunningPlugin? getRunningPlugin(String pluginId) => runningPlugins[pluginId];

  /// Inbox
  int selectedChannel = -1;
  List<Item> chats = [];
  List<Item> activeChatMessages = [];
  bool _loadingMessages = false;
  bool _loadingMoreMessages = false;
  bool _noChatRoomsFound = false;
  Map<String, String> chatImageSha = {};
  Map<String, ImageProvider> chatImageProviders = {};
  Map<String, String> messageImageSha = {};
  Map<String, String> tweetImageSha = {};
  Map<String, ImageProvider> messageImageProviders = {};
  Map<String, ImageProvider> tweetImageProviders = {};

  // ignore: cancel_subscriptions
  StreamSubscription<List<Item>>? chatStreamSubscription;

  // ignore: cancel_subscriptions
  StreamSubscription<List<Item>>? messageStreamSubscription;
  Stream<List<Item>>? chatStream, messageStream;
  String? selectedChatRoomID;
  ScrollController msgScrollController = ScrollController();

  bool get noChatRoomsFound => _noChatRoomsFound;

  bool get loadingMessages => _loadingMessages;

  bool get loadingMoreMessages => _loadingMoreMessages;

  Stream<List<Item>> gqlInboxChatsStream() async* {
    while (true) {
      await Future.delayed(Duration(milliseconds: 1500));
      List<Item> data = await _graphService.getInboxChats();
      _noChatRoomsFound = data.isEmpty;
      notifyListeners();
      if (data.isNotEmpty) {
        yield data;
      }
    }
  }

  Stream<List<Item>> gqlInboxMessagesStream(
      String chatID, int limit, int offset) async* {
    while (true) {
      await Future.delayed(Duration(milliseconds: 1500));
      List<Item> data =
          await _graphService.getInboxMessages(chatID, limit, offset);
      if (data.isNotEmpty) {
        yield data;
      }
    }
  }

  Stream<List<Item>> gqlStreamTwitterFeed() async* {
    while (true) {
      await Future.delayed(Duration(seconds: 1));
      List<Item> data = await _graphService.getTwitterFeed();
      if (data.isNotEmpty) {
        // if (activeFeedItem == null) activeFeedItem = items[0];
        yield data;
      }
    }
  }

  Stream<List<Item>>? appContentStream;
  StreamSubscription<Item>? appContentSubscription;
  List<Item> appContent = [];

  void getChats() async {
    chatStream = gqlInboxChatsStream();
    chatStreamSubscription = chatStream!.listen((res) async {
      // for (var item in res) {
      //   print(item.edgeItem("~messageChannel")?.get("dateSent"));
      // }
      sortByEdgeProperty(
          items: res,
          edge: '~messageChannel',
          property: 'dateSent',
          order: 'desc');
      for (var item in res) {
        var itemID = item.get('id');
        String? fileID =
            item.edgeItem('photo')?.edgeItem('file')?.get('sha256')?.toString();
        if (fileID != null &&
            (!chatImageSha.containsKey(itemID) ||
                (chatImageSha.containsKey(itemID) &&
                    chatImageSha[itemID] != fileID))) {
          chatImageSha[itemID] = fileID;
          chatImageProviders[itemID] = MemoryImage(await downloadPhoto(fileID));
        }
      }
      chats = res;
      notifyListeners();
    });
  }

  stopAppContentStream() {
    appContentSubscription?.cancel();
    appContentStream = null;
  }

  void setActiveChat({required String chatID, limit = 100, offset = 0}) async {
    _loadingMessages = true;
    notifyListeners();
    if (kDebugMode) {
      AppLogger.debug('offset: $offset');
    }
    if (offset == 0) {
      if (selectedChatRoomID == chatID) return;

      selectedChatRoomID = chatID;
      // Reset active messages while waiting for query
      if (messageStreamSubscription != null) {
        messageStreamSubscription!.cancel();
      }
      messageStream = null;
      activeChatMessages = [];
      messageImageSha = {};
      messageImageProviders = {};
    }

    List<Item> messages = [];

    if (offset == 0) {
      messageStream = gqlInboxMessagesStream(chatID, limit, offset);

      messageStreamSubscription = messageStream!.listen((result) async {
        if (result.isNotEmpty) {
          _loadingMessages = true;
          notifyListeners();
          messages = result[0].getEdgeTargets('~messageChannel') ?? [];
          bool exist = false;
          for (Item message in messages) {
            exist = false;
            for (Item activeMessage in activeChatMessages) {
              if (activeMessage.get('id') == message.get('id')) {
                exist = true;
                break;
              }
            }
            if (!exist) {
              activeChatMessages.add(message);
            }
          }
          await updateMessageImage(activeChatMessages);
          _loadingMessages = false;
          notifyListeners();
        }
      });
    } else {
      _loadingMessages = false;
      notifyListeners();
      List<Item> result =
          await _graphService.getInboxMessages(chatID, limit, offset);
      if (result.isNotEmpty) {
        messages = result[0].getEdgeTargets('~messageChannel') ?? [];
        if (messages.isEmpty) {
          _loadingMoreMessages = false;
          notifyListeners();
          return;
        }
        await updateMessageImage(messages);
        activeChatMessages.addAll(messages);
      }
      _loadingMoreMessages = false;
      notifyListeners();
    }
  }

  Future<void> updateMessageImage(List<Item> messages) async {
    for (var item in messages) {
      var itemID = item.get('id');
      if (messageImageProviders[itemID] != null) continue;
      String? fileID = item
          .edgeItem('sender')!
          .edgeItem('profilePicture')
          ?.edgeItem('file')
          ?.get('sha256')
          ?.toString();
      if (fileID != null &&
          (!messageImageSha.containsKey(itemID) ||
              (messageImageSha.containsKey(itemID) &&
                  messageImageSha[itemID] != fileID))) {
        messageImageSha[itemID] = fileID;
        messageImageProviders[itemID] =
            MemoryImage(await downloadPhoto(fileID));
      }
    }
  }

  void sortByEdgeProperty(
      {required List<Item> items,
      required String edge,
      required String property,
      required String order}) {
    if (order == 'asc') {
      items.sort((a, b) {
        if (a.edgeItem(edge)?.get(property) == null) {
          return 1;
        } else if (b.edgeItem(edge)?.get(property) == null) {
          return -1;
        }
        return a
            .edgeItem(edge)!
            .get(property)
            .compareTo(b.edgeItem(edge)!.get(property));
      });
    } else {
      items.sort((b, a) {
        if (a.edgeItem(edge)?.get(property) == null) {
          return 1;
        } else if (b.edgeItem(edge)?.get(property) == null) {
          return -1;
        }
        return a
            .edgeItem(edge)!
            .get(property)
            .compareTo(b.edgeItem(edge)!.get(property));
      });
    }
  }

  void handleOnTapSelectedChatRoom(int index) {
    _loadingMessages = true;
    notifyListeners();
    selectedChannel = index;
    String? chatID = chats[index].get('id')?.toString();

    setActiveChat(chatID: chatID!);
    msgScrollController = ScrollController();
    msgScrollController.addListener(msgScrollControllerListener);
    notifyListeners();
  }

  void onVisibilityChanged(VisibilityInfo info) {
    var visiblePercentage = info.visibleFraction * 100;
    if (visiblePercentage < 20) {
      if (chatStreamSubscription != null && !chatStreamSubscription!.isPaused) {
        chatStreamSubscription!.pause();
      }
      if (messageStreamSubscription != null &&
          !messageStreamSubscription!.isPaused) {
        messageStreamSubscription!.pause();
      }
    } else {
      if (chatStreamSubscription != null && chatStreamSubscription!.isPaused) {
        chatStreamSubscription!.resume();
      }
      if (messageStreamSubscription != null &&
          !messageStreamSubscription!.isPaused) {
        messageStreamSubscription!.resume();
      }
    }
  }

  void loadMoreMessage() {
    _loadingMoreMessages = true;
    notifyListeners();
    setActiveChat(
      chatID: selectedChatRoomID!,
      offset: activeChatMessages.length,
    );
  }

  void msgScrollControllerListener() {
    // take into consideration reversed order of chat messages ListView
    if (msgScrollController.offset >=
            msgScrollController.position.maxScrollExtent &&
        !msgScrollController.position.outOfRange) {
      loadMoreMessage();
    }
  }

  void disposeStreams() {
    chatStream = null;
    messageStream = null;
    if (chatStreamSubscription != null) chatStreamSubscription!.cancel();
    if (messageStreamSubscription != null) messageStreamSubscription!.cancel();
  }

  void getFeed() async {
    feedStreamSubscription = gqlStreamTwitterFeed().listen((res) {
      feed = res;
      setTwitterImages(feed);
      notifyListeners();
    });
  }

  void setTwitterImages(List<Item> tweets) async {
    for (var item in tweets) {
      var itemID = item.get('id');
      if (tweetImageProviders[itemID] != null) continue;
      String? fileID = item
          .edgeItem('author')!
          .edgeItem('profilePicture')
          ?.edgeItem('file')
          ?.get('sha256')
          ?.toString();
      if (fileID != null &&
          (!tweetImageSha.containsKey(itemID) ||
              (tweetImageSha.containsKey(itemID) &&
                  tweetImageSha[itemID] != fileID))) {
        tweetImageSha[itemID] = fileID;
        tweetImageProviders[itemID] = MemoryImage(await downloadPhoto(fileID));
      }
    }
  }
}
