import 'dart:async';
import 'dart:convert';

import 'package:flutter/widgets.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/services/gitlab_service.dart';
import 'package:memri/core/services/graph_service.dart';
import 'package:memri/core/services/pod_service.dart';
import 'package:memri/providers/oauth_provider.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/extensions/collection.dart';

class _Result {
  final List<Item> items;
  final List<Edge> edges;

  _Result(this.items, this.edges);
}

class ImporterProvider with ChangeNotifier {
  final PodService _podService;
  final GraphService _graphService;
  final OAuthProvider _oAuthProvider;
  final GitlabService _gitlabService;

  ImporterProvider(
    this._podService,
    this._graphService,
    this._oAuthProvider,
    this._gitlabService,
  );

  bool ableFetchImporters = true;
  Map<DataSource, Item> activePlugins = {};
  Timer? refreshActiveListTimer;

  Map<String, Stream<Item>?> downloadingStreams = {};
  Map<String, StreamSubscription<Item>> downloadingStreamSubs = {};
  Map<String, double?> progress = {};

  Map<DataSource, bool?> importerDataInPod = {};
  bool dataForLabelingInPod = false;

  Item? selectedPlugin;

  bool get isAnyImportersActive => activePlugins.values.isNotEmpty;

  bool isImporterActive(DataSource source) =>
      activePlugins[source] != null ||
      (importerDataInPod[source] != null && importerDataInPod[source]!) ||
      _oAuthProvider.oAuthCompletedByName(source);

  bool isImporterSelected(DataSource source) =>
      isImporterActive(source) &&
      selectedPlugin != null &&
      selectedPlugin == activePlugins[source];

  Future<void> fetchActiveImporters() async {
    if (!ableFetchImporters) return;
    ableFetchImporters = false;

    refreshActiveListTimer =
        Timer(Duration(seconds: 5), () => ableFetchImporters = true);

    var activeImporters = await _graphService.fetchActiveImporters();
    if (activeImporters.isNotEmpty) {
      var whatsapp = activeImporters.firstWhereOrNull((element) =>
          element.get('pluginName') == 'WhatsappPlugin' &&
          element.get('status') != 'error' &&
          element.get('status') != 'idle');
      if (whatsapp != null) activePlugins[DataSource.whatsapp] = whatsapp;
      var twitter = activeImporters.firstWhereOrNull((element) =>
          element.get('pluginName') == 'TwitterPlugin' &&
          element.get('status') != 'error' &&
          element.get('status') != 'idle');
      if (twitter != null) activePlugins[DataSource.twitter] = twitter;
      notifyListeners();
    }
  }

  void setActivePlugin(Item? plugin) {
    if (selectedPlugin == null || selectedPlugin != plugin) {
      selectedPlugin = plugin;
    } else {
      selectedPlugin = null;
    }
    notifyListeners();
  }

  void initDownloadingStreams(String id, String name) {
    downloadingStreams[name] = _downloadingItemStream(id);
    downloadingStreamSubs[name] = downloadingStreams[name]!.listen((item) {
      double? _progress = item.get('progress');

      if (_progress != null) {
        progress[name] = _progress;
        debugPrint('setting progress $_progress');
        if (_progress >= 1.0) {
          cancelDownloadingStreamSub(name);
        }
        notifyListeners();
      }
    });
  }

  Stream<Item> _downloadingItemStream(String id) async* {
    while (true) {
      await Future.delayed(Duration(seconds: 1));
      Item? res;

      var data = await _podService.getItem(id: id);
      var pluginRunItem = data;
      if (pluginRunItem != null) {
        res = pluginRunItem;
      }
      if (res != null) {
        yield res;
      }
    }
  }

  void cancelDownloadingStreamSub(String name) {
    downloadingStreams[name] = null;
    if (downloadingStreamSubs[name] != null) {
      downloadingStreamSubs[name]!.cancel();
    }
  }

  fetchImportersData() async {
    List<Item> messages = await _graphService.fetchWhatsappMessages();
    List<Item> tweets = await _graphService.fetchTwitterTweets();
    List<Item> sample_messages =
        await _graphService.fetchSampleWhatsappMessages();

    List<Item> items = messages + tweets + sample_messages;
    dataForLabelingInPod = items.isNotEmpty;
    importerDataInPod[DataSource.whatsapp] = messages.isNotEmpty;
    importerDataInPod[DataSource.twitter] = tweets.isNotEmpty;
    importerDataInPod[DataSource.sampleWhatsapp] = sample_messages.isNotEmpty;

    notifyListeners();
  }

  void importSampleData() {
    if (!(importerDataInPod[DataSource.sampleWhatsapp] ?? false)) {
      uploadSampleDataFromGitlab();
    }
  }

  Future<void> uploadSampleDataFromGitlab() async {
    final itemsToCreate = <Item>[];
    final edges = <Edge>[];

    final fileContent = await _gitlabService.getTextFileContent(
      projectPath: 'memri/sample-dataset',
      filepath: 'output/output.json',
      branch: 'main',
    );

    final json = jsonDecode(fileContent) as Map<String, dynamic>;
    for (final key in json.keys) {
      final incomingItems = json[key];
      final items =
          List<Item>.from(incomingItems.map((item) => Item.fromJson(item)));
      final result = await _createSampleChatFromItems(items);
      edges.addAll(result.edges);
      itemsToCreate.addAll(result.items);
    }

    debugPrint('Creating ${itemsToCreate.length} messages');
    await _podService.bulkAction(
        createItems: itemsToCreate, createEdges: edges);
  }

  Future<_Result> _createSampleChatFromItems(List<Item> items) async {
    final users = items.where((element) => element.type == 'Account');
    final channel =
        items.firstWhere((element) => element.type == 'MessageChannel');
    final messages = items.where((element) => element.type == 'Message');

    final createdUsers = List<Item>.from(users);

    var sender = createdUsers[0];
    var receiver = createdUsers[1];
    final messageEdges = <Edge>[];

    for (final message in messages) {
      messageEdges.addAll([
        Edge(source: message, target: channel, name: 'messageChannel'),
        Edge(source: message, target: sender, name: 'sender'),
        Edge(source: message, target: receiver, name: 'receiver'),
      ]);

      // swap sender and receiver
      final temp = sender;
      sender = receiver;
      receiver = temp;
    }

    final messagesList = messages.toList();
    messagesList.add(sender);
    messagesList.add(receiver);
    messagesList.add(channel);

    return _Result(messagesList, messageEdges);
  }

  void reset() {
    if (refreshActiveListTimer != null) {
      refreshActiveListTimer!.cancel();
    }
    activePlugins = {};
    importerDataInPod = {};
    progress = {};
    selectedPlugin = null;
    ableFetchImporters = true;
    dataForLabelingInPod = false;
  }

  refresh() {
    /// TODO update a state to handle it
    notifyListeners();
  }
}
