import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/services/pod_service.dart';
import 'package:memri/providers/project_provider.dart';

class LabellerProvider with ChangeNotifier {
  final ProjectProvider _projectProvider;
  final PodService _podService;
  Item? labellingTask;
  String errorMessage = '';

  LabellerProvider(this._podService, this._projectProvider);

  /*
  * Labeller Setup
  */
  void initLabellerSetup() {
    createLabellingTask();
  }

  void handleConfirmLabellerSetup(BuildContext context) async {
    if (getLabelOptions().length < 2) {
      errorMessage = 'A labelling task requires at least 2 label options';
    } else {
      await saveTaskToPod();
      RouteNavigator.navigateTo(
          context: context, route: Routes.projectsLabelData);
    }
  }

  void handleNewLabel(TextEditingController userInput) {
    if (userInput.text != '') {
      addLabelOption(userInput.text);
      userInput.text = '';
      notifyListeners();
    }
  }

  void handleRemoveLabelOption(int labelIndex) {
    // TODO track which edges to remove from pod if this is an existing task
    labellingTask!.edges['labelOption']!.targets.removeAt(labelIndex);
    notifyListeners();
  }

  /*
  * Labeller
  */

  // Save initial label to track if user changed label on next/previous/skip
  int? initialLabelIdx;
  int? selectedLabelIdx;
  int currentItemIdx = 0;
  bool showInfo = true;

  void handleSelectLabel(int selectionIdx) {
    // deselect label
    if (selectionIdx == selectedLabelIdx) {
      selectedLabelIdx = null;
    } else {
      selectedLabelIdx = selectionIdx;
    }
    handleLabelNext();
  }

  void handleLabelNext() {
    handleCreateLabel();
    // TODO remove this and add isSkipped to schema
    if (selectedLabelIdx == null) {
      numSkipped++;
    }

    currentItemIdx = (currentItemIdx + 1) % getDatasetSize();
    resetSelectedLabel();
    updateSummaryStatistics();
    notifyListeners();
  }

  void handleLabelPrevious() {
    handleCreateLabel();
    currentItemIdx = (currentItemIdx - 1) % getDatasetSize();
    resetSelectedLabel();
    updateSummaryStatistics();
    notifyListeners();
  }

  void handleCreateLabel() {
    // no changes
    if (initialLabelIdx == selectedLabelIdx) {
      return;
    }

    var entry = getDatasetEntry(currentItemIdx)!;

    String? selectedLabelValue;
    if (selectedLabelIdx != null) {
      selectedLabelValue = getLabelOptions()[selectedLabelIdx!].get('name');
    }

    // Remove existing edge if it exists
    List<Edge>? existingLabelEdges = entry.getEdges('annotation');
    if (existingLabelEdges != null && existingLabelEdges.isNotEmpty) {
      var existingLabelEdge = existingLabelEdges.first;
      entry.edges['annotation']?.targets.removeAt(0);
      _podService.removeEdge(existingLabelEdge);
      // TODO remove label item from Pod (endpoint not implemented)
    }

    // Add new label and edge
    if (selectedLabelIdx != null) {
      AppLogger.debug('Creating label: $selectedLabelIdx $selectedLabelValue');
      var label = Item(type: 'CategoricalLabel');
      label.properties['labelValue'] = selectedLabelValue;
      Edge labelEdge = entry.addEdge('annotation', label);
      _podService.bulkAction(createItems: [label], createEdges: [labelEdge]);
    }
  }

  void toggleShowInfo() {
    showInfo = false;
    notifyListeners();
  }

  /*
  * Methods
  */

  Item getLabel(int idx) {
    return getLabelOptions()[idx];
  }

  // String getCurrentContent() {
  //   var currentEntry = getDatasetEntry(currentItemIdx)!;
  //   String? content;
  //   var item = currentEntry.edges['data']?.first();
  //   // debugPrint(item.toString());
  //   // debugPrint("AA");
  //   if (_projectProvider.dataSourceSelected == DataSource.whatsapp){
  //     content = item?.get('content');
  //   }
  //   else{
  //     content = item?.get('message');
  //   }
  //   if (content == null) {
  //     return '<ERROR: CONTENT NOT FOUND>';
  //   } else {
  //     return content;
  //   }
  // }

  Item? getDatasetEntry(int idx) {
    return getDataset().getEdgeTargets('entry')?[idx];
  }

  void resetSelectedLabel() {
    // set selected and initial label idx to:
    // - null if current entry has no existing label
    // - idx of existing label if there is an existing label
    var entry = getDatasetEntry(currentItemIdx);
    var existingLabels = entry?.getEdgeTargets('annotation');
    if (existingLabels == null || existingLabels.isEmpty) {
      selectedLabelIdx = null;
      initialLabelIdx = null;
    } else {
      AppLogger.debug('${existingLabels.first.toJson()}');
      var existingLabelValue = existingLabels.first.get('labelValue');
      var existingLabelIdx = getLabelIndexFromValue(existingLabelValue)!;
      selectedLabelIdx = existingLabelIdx;
      initialLabelIdx = existingLabelIdx;
    }
  }

  int? getLabelIndexFromValue(String labelValue) {
    int idx = 0;
    for (var labelOption in getLabelOptions()) {
      if (labelOption.get('name') == labelValue) {
        return idx;
      } else {
        idx++;
      }
    }
    return null;
  }

  int getDatasetSize() {
    var entryEdge = getDataset().getEdgeTargets('entry');
    if (entryEdge == null) {
      return 0;
    } else {
      return entryEdge.length;
    }
  }

  void createLabellingTask() {
    var task = Item(type: 'LabellingTask');
    task.initEdge('labelOption');
    labellingTask = task;

    // Update dataset Item
    getDataset().addEdge('labellingTask', task);
  }

  Future<void> saveTaskToPod() async {
    // save task and labelOption edges + items
    var taskEdge = getDataset().getEdges('labellingTask')![0];
    var task = taskEdge.target;
    var labelOptionEdges = task.getEdges('labelOption')!;
    var labelOptions = getLabelOptions();

    var createItems = labelOptions + [task];
    var createEdges = labelOptionEdges + [taskEdge];

    AppLogger.debug(
        'Created labelTask with ${labelOptionEdges.length} label options.');

    return _podService.bulkAction(
        createItems: createItems, createEdges: createEdges);
  }

  Item getDataset() {
    return _projectProvider.currentDataset!;
  }

  List<Item> getLabelOptions() {
    return labellingTask?.getEdgeTargets('labelOption') ?? [];
  }

  void addLabelOption(String name) {
    Item labelOption = Item(type: 'LabelOption');
    labelOption.properties['name'] = name;
    labellingTask!.addEdge('labelOption', labelOption);
  }

  /*
  * Summary
  */
  int numLabeled = 0;
  int percentageLabeled = 0;
  int numSkipped = 0;
  int percentageSkipped = 0;
  Map<String, int> labelFrequencies = {};

  void updateSummaryStatistics() {
    numLabeled = 0;
    percentageLabeled = 0;
    labelFrequencies = {};
    for (int i = 0; i < getDatasetSize(); i++) {
      var annotations = getDatasetEntry(i)?.getEdgeTargets('annotation');
      if (annotations != null && annotations.isNotEmpty) {
        numLabeled++;
        String labelValue = annotations.first.get('labelValue');
        labelFrequencies[labelValue] = (labelFrequencies[labelValue] ?? 0) + 1;
      }
    }

    var datasetSize = getDatasetSize();
    if (datasetSize == 0) {
      percentageLabeled = 100;
      percentageSkipped = 100;
    } else {
      percentageLabeled =
          (numLabeled.toDouble() / datasetSize.toDouble() * 100).round();
      percentageSkipped =
          (numSkipped.toDouble() / datasetSize.toDouble() * 100).round();
    }
  }

  int getLabelPercentage(int? labelFreq) {
    if (numLabeled == 0 || labelFreq == null) {
      return 0;
    } else {
      return (labelFreq.toDouble() / numLabeled.toDouble() * 100).round();
    }
  }
}
