// ignore_for_file: cancel_subscriptions

import 'dart:async';
import 'dart:convert';
import 'dart:html' as html;
import 'dart:html';
import 'dart:js' as js;
import 'dart:math';
import 'dart:typed_data';

import 'package:crypto/crypto.dart';
import 'package:flutter/material.dart';
import 'package:http/http.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/services/analytics_service.dart';
import 'package:memri/core/services/gitlab_service.dart';
import 'package:memri/core/services/graph_service.dart';
import 'package:memri/core/services/oauth_service.dart';
import 'package:memri/core/services/pod_service.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:shared_preferences/shared_preferences.dart';

class OAuthProvider with ChangeNotifier {
  final SharedPreferences _prefs;
  final PodService _podService;
  final OAuthService _oAuthService;
  final GraphService _graphService;
  final GitlabService _gitlabService;

  OAuthProvider(
    this._prefs,
    this._podService,
    this._oAuthService,
    this._graphService,
    this._gitlabService,
  );

  String responseMessage = '';
  String errorMessage = 'Oops, something went wrong...!\nplease try again';
  String service = '';

  String? whatsappStatus;
  Uint8List? whatsappQrSvg;

  StreamSubscription<Item>? _whatsappOAuthStreamSubscription;
  StreamSubscription<dynamic>? _whatsappQrCodeStreamSubscription;
  Stream<Item>? _whatsappOAuthStream;
  Stream<dynamic>? _whatsappQrCodeStream;

  bool gitlabAuthFailed = false;
  Item? _gitlabOAuthItem;
  StreamSubscription<Item>? _gitlabOAuthStreamSubscription;
  Stream<Item>? _gitlabOAuthStream;

  Map<DataSource, bool> oAuthCompleted = {};

  String get endpoint => window.location.href.contains('localhost')
      ? 'http://localhost:3667'
      : app.settings.defaultAppUrl;

  String get oauthTokenSecret => _prefs.getString('oauthTokenSecret') ?? '';

  String get oauthCodeVerifier => _prefs.getString('oauthCodeVerifier') ?? '';

  set oauthTokenSecret(String val) => _prefs.setString('oauthTokenSecret', val);

  set oauthCodeVerifier(String val) =>
      _prefs.setString('oauthCodeVerifier', val);

  bool oAuthCompletedByName(DataSource dataSource) {
    if (oAuthCompleted.keys.isEmpty) return false;
    return oAuthCompleted[dataSource] ?? false;
  }

  void init() {
    var uri = Uri.parse(Uri.base.toString());
    var state = uri.queryParameters['state'];
    if (state != null && state.contains('twitter')) {
      service = 'twitter';
      handleTwitterOAuth(uri.queryParameters);
    } else {
      var code = uri.queryParameters['code'];
      var redirectUrl = '$endpoint/oauth';
      service = 'gitlab';
      if (code == null) {
        gitlabAuthFailed = true;
        responseMessage = errorMessage;
        AppLogger.info('No code provided');
      } else {
        _requestGitlabTokens(code, redirectUrl);
      }
    }
  }

  Future<dynamic> _requestGitlabTokens(String code, String redirectUrl) async {
    String? codeVerifier = oauthCodeVerifier;
    String clientId =
        'b87266bd2290a9992ea28ff4a6b4f27d59603b6e67f4b35259a098bfa7135898';
    final queryParameters = {
      'client_id': clientId,
      'code': code,
      'grant_type': 'authorization_code',
      'redirect_uri': redirectUrl,
      'code_verifier': codeVerifier,
    };
    var uri = Uri.https('gitlab.memri.io', '/oauth/token', queryParameters);
    Response response = await post(uri);
    var resBody = jsonDecode(response.body);

    //success
    if (resBody['access_token'] != null) {
      String accessToken = resBody['access_token']!;
      String refreshToken = resBody['refresh_token']!;
      createOAuthItem('gitlab', accessToken, refreshToken);
      finishOAuth('GitLab');
    }
    //failure
    else {
      responseMessage = 'Oops, something went wrong. please try again';
      notifyListeners();
      AppLogger.err(response.toString());
    }
  }

  Future<Item?> getOAuthItem(String service) async {
    var items = await _graphService.getOAuthFlowItems(service);
    if (items.isNotEmpty) {
      items.sort((a, b) =>
          (a.get('dateCreated') as num).compareTo(b.get('dateCreated') as num));
      return items.last;
    } else {
      return null;
    }
  }

  void createOAuthItem(
      String service, String oauthToken, String oauthTokenSecret) {
    _podService.createItem(
        item: Item.fromJson({
      'type': 'OauthFlow',
      'service': service,
      'accessToken': oauthToken,
      'accessTokenSecret': oauthTokenSecret,
    }));
  }

  void openOAuthWindow() {
    const _chars = '123456789';
    Random _rnd = Random();

    String getRandomString(int length) =>
        String.fromCharCodes(Iterable.generate(
            length, (_) => _chars.codeUnitAt(_rnd.nextInt(_chars.length))));
    String oauthCodeVerifier = getRandomString(100);
    _prefs.setString('oauthCodeVerifier', oauthCodeVerifier);
    List<int> bytes = utf8.encode(oauthCodeVerifier);
    List<int> hash = sha256.convert(bytes).bytes;
    String codeChallenge = base64UrlEncode(hash);
    if (codeChallenge[codeChallenge.length - 1] == '=') {
      codeChallenge = codeChallenge.substring(0, codeChallenge.length - 1);
    }

    String redirectUrl = '$endpoint/oauth';
    String state = getRandomString(128);
    String scopes = 'api read_repository write_repository';
    String clientId =
        'b87266bd2290a9992ea28ff4a6b4f27d59603b6e67f4b35259a098bfa7135898';

    final queryParameters = {
      'client_id': clientId,
      'response_type': 'code',
      'grant_type': 'authorization_code',
      'redirect_uri': redirectUrl,
      'state': state,
      'scope': scopes,
      'code_challenge': codeChallenge,
      'code_challenge_method': 'S256'
    };

    var url = Uri.https('gitlab.memri.io', '/oauth/authorize', queryParameters);
    html.window
        .open(url.toString(), '_blank', 'location=yes, height=800,width=600');
  }

  Future<Map<String, dynamic>> _requestOauthAccessToken(String oauthVerifier,
          String oauthToken, String oauthTokenSecret) async =>
      await _oAuthService.oauthAccessToken(
          oauthVerifier, oauthToken, oauthTokenSecret);

  _setOAuthCompleted(bool val) async {
    await _prefs.setBool('oauthCompleted', val);
    notifyListeners();
  }

  /// GITLAB
  bool _gitlabUserExist = false;

  bool get gitlabOAuthExist => _gitlabOAuthItem != null && _gitlabUserExist;

  Future<void> initGitlabOAuth() async {
    await setGitlabTokens();
    if (_gitlabOAuthItem == null) {
      _gitlabOAuthStream = _gitlabItemStream();
      _gitlabOAuthStreamSubscription = _gitlabOAuthStream!.listen((item) {
        _gitlabOAuthItem = item;
        if (item.get('accessToken') != null) {
          setGitlabTokens();
          _gitlabUserExist = true;
          resetGitlabStream();
        }
        notifyListeners();
      });
    } else {
      await _gitlabService.getUser().then((value) {
        _gitlabUserExist = true;
        notifyListeners();
      }).catchError((error) {
        _gitlabUserExist = false;
        _gitlabOAuthItem = null;
        notifyListeners();
      });
    }
  }

  Stream<Item> _gitlabItemStream() async* {
    while (true) {
      await Future.delayed(Duration(seconds: 1));
      final gitlabItem = await getOAuthItem('gitlab');
      if (gitlabItem != null) {
        // HACK: We ask user to authorise again if the token is expired
        // as a workaround for not using refresh tokens
        final msSinceCreate = DateTime.now().millisecondsSinceEpoch -
            gitlabItem.get('dateCreated');
        final minsSinceCreate = msSinceCreate ~/ 1000 ~/ 60;
        const minsToExpire = 120;

        if (minsSinceCreate < minsToExpire) yield gitlabItem;
      }

      /// todo: refresh token if needed}
    }
  }

  Future<void> setGitlabTokens() async {
    _gitlabOAuthItem ??= await getOAuthItem('gitlab');
    if (_gitlabOAuthItem != null) {
      var accessToken = _gitlabOAuthItem!.get('accessToken');
      var refreshToken = _gitlabOAuthItem!.get('refreshToken');

      if (accessToken != null) {
        _gitlabService.setTokens(accessToken, refreshToken);
        _gitlabUserExist = true;
      } else {
        AppLogger.err('Could not setup GitLab. Token is expired.');
      }
    } else {
      _gitlabUserExist = false;
    }
  }

  resetGitlabStream() {
    _gitlabOAuthStream = null;
    if (_gitlabOAuthStreamSubscription != null) {
      _gitlabOAuthStreamSubscription!.cancel();
    }
  }

  /// Twitter

  Future<void> initTwitterOAuth(BuildContext context) async {
    String redirectUrl = '$endpoint/oauth?state=twitter';
    _setOAuthCompleted(false);

    var res = await _oAuthService.oauthRequestToken('twitter', redirectUrl);
    final queryParameters = {'oauth_token': res['oauth_token']};

    // we are using this in the redirect screen
    _prefs.setString('oauthTokenSecret', res['oauth_token_secret']);

    var url = Uri.https('api.twitter.com', '/oauth/authorize', queryParameters);

    html.window
        .open(url.toString(), '_blank', 'location=yes, height=800,width=600');

    Timer.periodic(Duration(seconds: 1), (timer) async {
      _prefs.reload();
      notifyListeners();

      oAuthCompleted[DataSource.twitter] =
          _prefs.getBool('oauthCompleted') ?? false;

      if (oAuthCompleted[DataSource.twitter]!) {
        AppLogger.info('oauth completed');
        Item item = await _podService.startImporter('TwitterPlugin',
            'twitter.plugin', app.settings.twitterContainer, 'twitter');
        var status = item.get('status');
        notifyListeners();
        timer.cancel();
        if (status == 'error') {
          ///
          /// TODO
          ///
        } else {
          String id = item.id;
          RouteNavigator.navigateTo(
            context: context,
            route: Routes.importerDownloading,
            param: {'id': id, 'name': 'twitter'},
            replace: true,
          );
        }
      }
    });
  }

  void handleTwitterOAuth(Map<String, String> previousQueryParameters) async {
    var resBody = await _requestOauthAccessToken(
        previousQueryParameters['oauth_verifier']!,
        previousQueryParameters['oauth_token']!,
        oauthTokenSecret);

    //success
    if (resBody['oauth_token'] != null) {
      String oauthToken = resBody['oauth_token']!;
      String oauthTokenSecret = resBody['oauth_token_secret']!;
      createOAuthItem('twitter', oauthToken, oauthTokenSecret);
      AppLogger.info('setting');

      _setOAuthCompleted(true);
      finishOAuth('Twitter');
    }
    //failure
    else {
      responseMessage = 'Oops, something went wrong. please try again';
      notifyListeners();
      AppLogger.err(resBody.toString());
    }
    // close screen
  }

  void finishOAuth(String title) async {
    for (var i in [3, 2, 1]) {
      await Future.delayed(Duration(seconds: 1));
      responseMessage =
          'Successfully authenticated $title\nRedirecting you back in $i seconds';
      notifyListeners();
    }
    await Future.delayed(Duration(seconds: 1));
    js.context.callMethod('close');
  }

  /// WhatsApp

  void initWhatsappOauth(BuildContext context) async {
    // TODO, what are we doing with things from the default database?

    try {
      Item item = await _podService.startImporter('WhatsappPlugin',
          'whatsapp.plugin', app.settings.whatsappContainer, 'whatsapp');
      String id = item.id;
      service = 'whatsapp';
      _whatsappOAuthStream ??= _whatsappItemStream(id);
      _whatsappOAuthStreamSubscription =
          _whatsappOAuthStream!.listen((item) async {
        cancelWhatsappStreamSubs();
        String? webserverUrl = item.get('webserverUrl');
        if (webserverUrl != null) {
          _whatsappQrCodeStream ??= _whatsappQrStream(webserverUrl);
          _whatsappQrCodeStreamSubscription ??=
              _whatsappQrCodeStream!.listen((dynamic res) async {
            if (res != null) {
              oAuthCompleted[DataSource.whatsapp] = res['authenticated'];
              if (oAuthCompleted[DataSource.whatsapp]!) {
                RouteNavigator.navigateTo(
                  context: context,
                  route: Routes.importerDownloading,
                  param: {'id': id, 'name': 'whatsapp'},
                  replace: true,
                );
              } else {
                String? svg = res['qr'];
                if (svg != null) {
                  whatsappQrSvg = base64.decode(svg.split(',').last);
                  notifyListeners();
                }
              }
            }
          });

          var status = item.get('status');
          if (status != whatsappStatus) {
            AnalyticsService().logImporterStatus(status);
          }
          whatsappStatus = status;
          AppLogger.info(whatsappStatus);

          notifyListeners();
        }
      });
    } catch (error) {
      AppLogger.err('Error starting plugin: $error');
    }
  }

  void cancelWhatsappStreamSubs() {
    _whatsappOAuthStream = null;
    if (_whatsappOAuthStreamSubscription != null) {
      _whatsappOAuthStreamSubscription!.cancel();
    }
  }

  void cancelWhatsappQrCodeStreamSubs() {
    _whatsappQrCodeStream = null;
    if (_whatsappQrCodeStreamSubscription != null) {
      _whatsappQrCodeStreamSubscription!.cancel();
    }
  }

  Stream<Item> _whatsappItemStream(String id) async* {
    while (true) {
      await Future.delayed(Duration(seconds: 1));
      Item? res;
      Item? pluginRunItem = await _podService.getItem(id: id);
      if (pluginRunItem != null) {
        res = pluginRunItem;
      }

      if (res != null) {
        yield res;
        if (res.get('webserverUrl') != null) {
          cancelWhatsappStreamSubs();
        }
      }
    }
  }

  Stream<dynamic> _whatsappQrStream(String webserverUrl) async* {
    while (true) {
      await Future.delayed(Duration(seconds: 1));
      dynamic res = await _oAuthService.getQrCodeSvg(webserverUrl);
      if (res != null) {
        yield res;
        if (res['authenticated']) {
          cancelWhatsappQrCodeStreamSubs();
        }
      }
    }
  }

  ///

  void reset() {
    cancelWhatsappStreamSubs();
    cancelWhatsappQrCodeStreamSubs();
    _gitlabOAuthStream = null;
    _gitlabOAuthItem = null;
    whatsappQrSvg = null;
    whatsappStatus = null;
    oAuthCompleted = {};
    gitlabAuthFailed = false;
    service = '';
    responseMessage = '';
  }
}
