// ignore_for_file: cancel_subscriptions

import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/models/pipeline_job.dart';
import 'package:memri/core/services/analytics_service.dart';
import 'package:memri/core/services/gitlab_service.dart';
import 'package:memri/core/services/graph_service.dart';
import 'package:memri/core/services/pod_service.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/widgets/dialogs/custom_alert_dialog.dart';
import 'package:uuid/uuid.dart';

enum DataSource { twitter, whatsapp, sampleWhatsapp }

class ProjectProvider with ChangeNotifier {
  final PodService _podService;
  final GraphService _graphService;
  final GitlabService _gitlabService;

  ProjectProvider(this._podService, this._graphService, this._gitlabService);

  List<String> notSubmittedLabelValues = [];

  String errorMessage = '';
  bool whatsappSelected = false;
  DataSource? dataSourceSelected;

  Map<DataSource, List<String>> features = {
    DataSource.whatsapp: ['Message Text', 'Data Sent'],
    DataSource.twitter: ['Tweet Text', 'Author', 'Post Date'],
    DataSource.sampleWhatsapp: ['Message Text', 'Data Sent']
  };

  List<Item> projects = [];

  Map<String, DatasetSummary> datasetSummaries = {};
  Item? currentProject;
  Item? currentDataset;
  Item? labellingTask;

  // Gitlab state
  bool projectHasPackage = false;
  bool projectCIPassed = false;
  List<PipelineJob> projectPipelines = [];
  Map<String, dynamic> gitlabProject = {};

  Stream<bool>? projectPackageStream;
  Stream<List<PipelineJob>>? projectPipelineStream;
  StreamSubscription<bool>? projectPackageSubscription;
  StreamSubscription<List<PipelineJob>>? projectPipelineSubscription;

  List<String> selectedFeatures = [];

  reset() {
    _gitlabService.setTokens(null, null);
    notSubmittedLabelValues = [];
    selectedFeatures = [];
    whatsappSelected = false;
    dataSourceSelected = null;

    resetProjectState();

    projectPackageStream = null;
    projectPipelineStream = null;
    projectHasPackage = false;
    projectCIPassed = false;
    projectPipelines = [];
    gitlabProject = {};
    datasetSummaries = {};
  }

  resetProjectState() {
    // Reset provider current project state
    errorMessage = '';
    whatsappSelected = false;
    dataSourceSelected = null;
    projects = [];
    currentProject = null;
    currentDataset = null;
    labellingTask = null;
    resetLabeller();
  }

  handleError(BuildContext context) {
    WidgetsBinding.instance!.addPostFrameCallback((_) =>
        RouteNavigator.navigateTo(context: context, route: Routes.error));
  }

  Future<void> setCurrentProject(BuildContext context, String projectId,
      {bool resetState = true}) async {
    // Reset provider state, and get project, dataset, labellerTask from pod
    if (projectId.isEmpty) {
      handleError(context);
      return;
    }

    // no changes if current project does not change
    if (projectId == currentProject?.get('id')) {
      return;
    }

    if (resetState) resetProjectState();

    var result = await _graphService.getProjectItems(projectId);
    if (result.isEmpty) {
      handleError(context);
      return;
    }

    // TODO handle missing dataset or labellingTask when loading project from Pod
    AppLogger.debug("Setting project: ${result[0].get("id")}");

    currentProject = result.first;
    currentDataset = currentProject!.edgeItem('dataset');
    labellingTask = currentDataset?.edgeItem('labellingTask');
    updateCurrentSummary();
    notifyListeners();
  }

  void deleteDataset(BuildContext context, Item dataset) {
    showDialog(
      context: context,
      builder: (context) => CustomAlertDialog(
        title: 'Delete ${dataset.get("name")}',
        description: 'Are you sure you want to delete this dataset?',
        button: TextButton(
          child: Text('Delete'),
          onPressed: () {
            RouteNavigator.navigateBack(context);
            _podService.deleteItem(dataset.get('id'));
            notifyListeners();
          },
        ),
      ),
    );
  }

  void deleteProject(BuildContext context, Item project) {
    showDialog(
      context: context,
      builder: (context) => CustomAlertDialog(
        title: 'Delete ${project.get("name")}',
        description: 'Are you sure you want to delete this project?',
        button: TextButton(
          child: Text('Delete'),
          onPressed: () {
            RouteNavigator.navigateBack(context);
            RouteNavigator.navigateBack(context);
            projects.removeWhere((element) => element.id == project.id);
            _podService.deleteItem(project.get('id'));
            notifyListeners();
          },
        ),
      ),
    );
  }

  Future<bool> projectNameExists(String projectName) async {
    var projects = await _graphService.getAllProjectsNames();
    var existingProjectNames = projects.map((p) => p.get('name')).toSet();
    return !existingProjectNames.contains(projectName);
  }

  getProjectSummaries() async {
    projects = await _graphService.getProjectSummaries();

    for (Item project in projects) {
      var summary = DatasetSummary(dataset: project.edgeItem('dataset'));
      datasetSummaries[project.get('id')!] = summary;
    }

    notifyListeners();
  }

  handleSelectDataSource(DataSource dataSource, {bool notify = true}) {
    if (dataSource == dataSourceSelected) {
      dataSourceSelected = null;
    } else {
      dataSourceSelected = dataSource;
    }
    if (notify) notifyListeners();
    AnalyticsService().logProjectDataSelect(dataSource);
  }

  handleCreateProject(BuildContext context, String projectName) async {
    if (projectName.isEmpty) {
      errorMessage = 'Enter a project name';
      notifyListeners();
    } else if (!_gitlabService.isValidGitlabSlug(projectName)) {
      // TODO have a separate field for gitlab projectname and displayname
      errorMessage =
          'Your project name should only contain lowercase letters, numbers, or dashes';
      notifyListeners();
    } else if (!await projectNameExists(projectName)) {
      errorMessage = 'Project name already exists';
      notifyListeners();
    } else if (dataSourceSelected == null) {
      errorMessage = 'Select a datasource first';
      notifyListeners();
    } else if (notSubmittedLabelValues.length < 2) {
      errorMessage = 'Enter at least 2 label options';
      notifyListeners();
    } else {
      // TODO save everything in single bulk request
      // TODO change the whatsapp to something dynamic
      AnalyticsService().logProjectCreate(projectName, 'whatsapp');
      await createProject(name: projectName);
      await createDataset();
      AnalyticsService().logProjectAddLabels(notSubmittedLabelValues);
      await createLabellingTask();
      await createGitlabProject(repo: projectName);
      notSubmittedLabelValues = [];
      RouteNavigator.navigateTo(
          context: context,
          route: Routes.projectsLabelData,
          param: {'id': currentProject!.get('id')});
    }
  }

  handleCreateDataset(BuildContext context, List<String> properties,
      List<bool> selected) async {
    AppLogger.debug('CREATING DATASET');
    await createDataset();
    RouteNavigator.navigateTo(
        context: context, route: Routes.projectsSetupLabelEditor);
  }

  createGitlabProject({required String repo}) async {
    AppLogger.debug('CREATING GITLAB PROJECT');
    String id = Uuid().v4();
    Item item = Item(
      type: 'PluginRun',
      properties: {
        'id': id,
        'pluginName': 'FactoryPlugin',
        'pluginModule': 'plugin_factory.plugin',
        'containerImage': app.settings.pluginFactoryContainer,
        'status': 'idle',
        'targetItemId': id,
        'config': '{"repo": "$repo"}',
      },
    );

    List<Item> createItems = [item];
    await _podService.bulkAction(createItems: createItems);
  }

  createProject({required String name}) async {
    var project = Item(type: 'Project');
    project.properties['name'] = name;
    await _podService.createItem(item: project);
    currentProject = project;
    AppLogger.debug('created project: ${project.toJson()}');
  }

  createDataset() async {
    var dataset = await _podService.createDataset(currentProject!);
    currentProject!.edges['dataset'] =
        EdgeList(name: 'dataset', targets: [dataset]);

    List<Item> datasetItems;
    List<Item> datasetFeatures = [];
    switch (dataSourceSelected) {
      case DataSource.twitter:
        datasetItems = await _graphService.getTwitterFeed();
        Item? feature =
            await _graphService.getPropertySchema('Tweet', 'message');
        datasetFeatures.add(feature!);
        break;
      case DataSource.whatsapp:
        datasetItems = await _graphService.getMessages('whatsapp');
        Item? feature =
            await _graphService.getPropertySchema('Message', 'content');
        datasetFeatures.add(feature!);
        break;
      case DataSource.sampleWhatsapp:
        datasetItems = await _graphService.getMessages('sample-whatsapp');
        Item? feature =
            await _graphService.getPropertySchema('Message', 'content');
        datasetFeatures.add(feature!);
        break;
      default:
        datasetItems = [];
        break;
    }

    List<Edge> featureEdges = [];
    for (Item feature in datasetFeatures) {
      Edge featureEdge = dataset.addEdge('feature', feature);
      featureEdges.add(featureEdge);
    }

    List<int> indices = [];
    if (datasetItems.isNotEmpty) {
      AppLogger.debug('creating dataset with length ${datasetItems.length}');
      indices = List<int>.generate(datasetItems.length - 1, (i) => i + 1);
      indices.shuffle();
    }

    List<Item> entries = [];
    List<Edge> dataEdges = [];
    List<Edge> entryEdges = [];
    for (var index in indices) {
      var item = datasetItems[index];
      Item entry = Item.fromJson({'type': 'DatasetEntry'});
      var dataEdge = entry.addEdge('data', item);
      var entryEdge = dataset.addEdge('entry', entry);
      entries.add(entry);
      dataEdges.add(dataEdge);
      entryEdges.add(entryEdge);
    }

    await _podService.bulkAction(
      createEdges: dataEdges + entryEdges + featureEdges,
      createItems: entries,
    );

    AppLogger.debug('created dataset with ${entryEdges.length} entries');
    currentDataset = dataset;
  }

  /*
  * Setup Labeller
  */

  void handleNewLabel(TextEditingController userInput, FocusNode focusNode) {
    if (userInput.text != '') {
      notSubmittedLabelValues.add(userInput.text);
      userInput.text = '';
      focusNode.requestFocus();
      notifyListeners();
    }
  }

  void handleRemoveLabelOption(int labelIndex) {
    notSubmittedLabelValues.removeAt(labelIndex);
    notifyListeners();
  }

  /*
  * Labeller
  */

  // Save initial label to track if user changed label on next/previous/skip
  int? initialLabelIdx;
  int? selectedLabelIdx;
  int currentItemIdx = 0;
  bool showInfo = true;

  bool isCurrentLabelSelected(int index) => selectedLabelIdx == index;

  resetLabeller() {
    initialLabelIdx = null;
    selectedLabelIdx = null;
    currentItemIdx = 0;
    showInfo = true;
    currentSummary.reset();
  }

  void handleResumeSetup(BuildContext context) {
    if (currentProject != null) {
      showDialog(
        context: context,
        builder: (context) => CustomAlertDialog(
          title: 'Caution',
          description:
              'If you go back to your project setup, your labeling progress will reset',
          button: TextButton(
            child: Text('Back to labeling setup'),
            onPressed: () {
              RouteNavigator.navigateBack(context);
              RouteNavigator.navigateBack(context);
              projects
                  .removeWhere((element) => element.id == currentProject!.id);
              _podService.deleteItem(currentProject!.get('id'));
              resetProjectState();
              notifyListeners();
            },
          ),
        ),
      );
    }
  }

  void handleSelectLabel(int selectionIdx) {
    // deselect label
    if (selectionIdx == selectedLabelIdx) {
      selectedLabelIdx = null;
    } else {
      selectedLabelIdx = selectionIdx;
    }
    notifyListeners();
  }

  void handleLabelNext() {
    handleCreateLabel();
    handleSkipLabel();
    currentItemIdx = (currentItemIdx + 1) % getDatasetSize();
    resetSelectedLabel();
    updateCurrentSummary();
    notifyListeners();
  }

  void handleLabelPrevious() {
    handleCreateLabel();
    currentItemIdx = (currentItemIdx - 1) % getDatasetSize();
    resetSelectedLabel();
    updateCurrentSummary();
    notifyListeners();
  }

  void handleLabelKeyboardEvents(KeyEvent event) {
    var pressedKey = int.tryParse(event.character ?? '');
    if (pressedKey is int) {
      var index = pressedKey - 1;
      if (index < 0) {
        selectedLabelIdx = 0;
      } else if (getLabelOptions().length > index) {
        selectedLabelIdx = index;
      }
    } else if (event is KeyUpEvent) {
      if (event.logicalKey == LogicalKeyboardKey.arrowDown) {
        if (selectedLabelIdx == null) {
          selectedLabelIdx = 0;
        } else {
          if (getLabelOptions().length - 1 != selectedLabelIdx) {
            selectedLabelIdx = selectedLabelIdx! + 1;
          }
        }
      } else if (event.logicalKey == LogicalKeyboardKey.arrowUp) {
        if (selectedLabelIdx == null) {
          selectedLabelIdx = getLabelOptions().length - 1;
        } else {
          if (selectedLabelIdx! <= 1) {
            selectedLabelIdx = 0;
          } else {
            selectedLabelIdx = selectedLabelIdx! - 1;
          }
        }
      } else if (event.logicalKey == LogicalKeyboardKey.arrowRight) {
        handleLabelNext();
      } else if (event.logicalKey == LogicalKeyboardKey.arrowLeft) {
        handleLabelPrevious();
      } else if (event.logicalKey == LogicalKeyboardKey.enter) {
        if (selectedLabelIdx != null) {
          handleLabelNext();
        }
      }
    }
    notifyListeners();
  }

  void handleSkipLabel() {
    if (selectedLabelIdx == null) {
      var entry = _currentItemDatasetEntry!;
      entry.properties['skippedByLabeller'] = true;
      _podService.bulkAction(updateItems: [entry]);
    }
  }

  void handleCreateLabel() {
    var entry = _currentItemDatasetEntry!;

    // no changes
    if (initialLabelIdx == selectedLabelIdx) {
      return;
    }

    String? selectedLabelValue;
    if (selectedLabelIdx != null) {
      selectedLabelValue = getLabelOptions()[selectedLabelIdx!].get('name');
    }

    // Remove existing edge if it exists
    List<Edge>? existingLabelEdges = entry.getEdges('annotation');
    if (existingLabelEdges != null && existingLabelEdges.isNotEmpty) {
      var existingLabelEdge = existingLabelEdges.first;
      entry.edges['annotation']?.targets.removeAt(0);
      _podService.removeEdge(existingLabelEdge);
      _podService.deleteItem(existingLabelEdge.target.get('id'));
    }

    // Add new label and edge
    if (selectedLabelIdx != null) {
      AppLogger.debug('Creating label: $selectedLabelIdx $selectedLabelValue');
      var label = Item(type: 'CategoricalLabel');
      label.properties['labelValue'] = selectedLabelValue;
      Edge labelEdge = entry.addEdge('annotation', label);
      entry.properties['skippedByLabeller'] = false;
      _podService.bulkAction(
          createItems: [label], createEdges: [labelEdge], updateItems: [entry]);
    }
  }

  void toggleShowInfo() {
    showInfo = false;
    notifyListeners();
  }

  /*
  * Create app
  */

  void handleCreateApp(BuildContext context) {
    if (projectCIPassed && projectHasPackage) {
      if (projectPackageSubscription != null) {
        projectPackageSubscription!.cancel();
      }
      if (projectPipelineSubscription != null) {
        projectPipelineSubscription!.cancel();
      }
      projectPackageStream = null;
      projectPipelineStream = null;
      AnalyticsService().logProjectTrainModel();
      String? projectId = currentProject?.get('id');
      RouteNavigator.navigateTo(
          route: Routes.projectsAppPreview,
          context: context,
          param: {'id': projectId});
    }
  }

  /*
  * Methods
  */

  Item getLabel(int idx) {
    return getLabelOptions()[idx];
  }

  String getCurrentContent() {
    var currentEntry = _currentItemDatasetEntry;
    var item = currentEntry?.edges['data']?.first();
    String content = '';

    switch (dataSourceSelected) {
      case DataSource.twitter:
        content = item?.get('message');
        break;
      case DataSource.whatsapp:
      case DataSource.sampleWhatsapp:
        content = item?.get('content') ?? '<CONTENT NOT FOUND>';
        break;
      default:
        content = '<CONTENT NOT FOUND>';
        break;
    }

    return content;
  }

  Item? get _currentItemDatasetEntry =>
      currentDataset?.getEdgeTargets('entry')?[currentItemIdx];

  void resetSelectedLabel() {
    // set selected and initial label idx to:
    // - null if current entry has no existing label
    // - idx of existing label if there is an existing label
    var entry = _currentItemDatasetEntry;
    var existingLabels = entry?.getEdgeTargets('annotation');
    if (existingLabels == null || existingLabels.isEmpty) {
      selectedLabelIdx = null;
      initialLabelIdx = null;
    } else {
      AppLogger.debug('${existingLabels.first.toJson()}');
      var existingLabelValue = existingLabels.first.get('labelValue');
      var existingLabelIdx = getLabelIndexFromValue(existingLabelValue)!;
      selectedLabelIdx = existingLabelIdx;
      initialLabelIdx = existingLabelIdx;
    }
  }

  int? getLabelIndexFromValue(String labelValue) {
    int idx = 0;
    for (var labelOption in getLabelOptions()) {
      if (labelOption.get('name') == labelValue) {
        return idx;
      } else {
        idx++;
      }
    }
    return null;
  }

  int getDatasetSize() {
    List<Item>? entryEdge;
    if (getDataset() != null) entryEdge = getDataset()!.getEdgeTargets('entry');
    if (entryEdge == null) {
      return 0;
    } else {
      return entryEdge.length;
    }
  }

  createLabellingTask() async {
    var task = Item(type: 'LabellingTask');
    task.initEdge('labelOption');
    labellingTask = task;
    for (String labelValue in notSubmittedLabelValues) {
      addLabelOption(labelValue);
    }

    getDataset()!.addEdge('labellingTask', task);
    await saveTaskToPod();
  }

  saveTaskToPod() async {
    // save task and labelOption edges + items
    var taskEdge = getDataset()!.getEdges('labellingTask')![0];
    var task = taskEdge.target;
    var labelOptionEdges = task.getEdges('labelOption')!;
    var labelOptions = getLabelOptions();

    var createItems = labelOptions + [task];
    var createEdges = labelOptionEdges + [taskEdge];

    AppLogger.debug(
        'Created labelTask with ${labelOptionEdges.length} label options.');

    return _podService.bulkAction(
        createItems: createItems, createEdges: createEdges);
  }

  Item? getDataset() {
    return currentDataset;
  }

  List<Item> getLabelOptions() {
    return labellingTask?.getEdgeTargets('labelOption') ?? [];
  }

  void addLabelOption(String name) {
    Item labelOption = Item(type: 'LabelOption');
    labelOption.properties['name'] = name;
    labellingTask!.addEdge('labelOption', labelOption);
  }

  /*
  * Dataset Summary
  */

  DatasetSummary currentSummary = DatasetSummary();

  updateCurrentSummary() {
    if (currentDataset != null) {
      currentSummary.dataset = currentDataset;
      currentSummary.update();
    }
    notifyListeners();
  }

  int getLabelPercentage(int? labelFreq) {
    return currentSummary.getLabelPercentage(labelFreq);
  }

  /*
  * Project Summary
  */

  handleDeleteProject(BuildContext context) {
    if (currentProject != null) {
      deleteProject(context, currentProject!);
    }
  }

  handleDeleteDataset(BuildContext context) {
    if (currentDataset != null) {
      deleteDataset(context, currentDataset!);
    }
  }

  setGitlabProject() async {
    // TODO Error handling
    String? projectName = currentProject?.get('name');
    if (projectName != null) {
      gitlabProject = await _gitlabService.getUserProject(projectName);
      AppLogger.debug(
          'set gitlab project ${gitlabProject['id']} ${gitlabProject['web_url']}');
      notifyListeners();
    }
  }

  void setupProjectPackageStream() {
    projectPackageStream = _checkProjectPackageStream();
    projectPackageSubscription = projectPackageStream!.listen((hasPackage) {
      AppLogger.debug('checking model package: $hasPackage');
      if (hasPackage != projectHasPackage) {
        projectHasPackage = hasPackage;
        if (hasPackage) {
          if (projectPackageSubscription != null) {
            projectPackageSubscription!.cancel();
          }
          projectPackageStream = null;
        }
        notifyListeners();
      }
    });
  }

  Stream<bool> _checkProjectPackageStream() async* {
    while (true) {
      await Future.delayed(Duration(seconds: 3));
      if (_gitlabService.accessToken != null && gitlabProject['id'] != null) {
        yield await _gitlabService.projectHasModelPackage(gitlabProject['id']!);
      }
    }
  }

  void setupProjectPipelineStream() {
    projectPipelineStream = _projectPipelineStream();
    projectPipelineSubscription = projectPipelineStream!.listen((pipelines) {
      AppLogger.debug('found project pipelines: ${pipelines.length}');
      if (pipelines.isNotEmpty) {
        projectPipelines = pipelines;

        var allSuccess = true;
        for (var pipeline in projectPipelines) {
          if (pipeline.status != 'success') {
            allSuccess = false;
            break;
          }
        }

        projectCIPassed = allSuccess;
        notifyListeners();
      }
    });
  }

  Stream<List<PipelineJob>> _projectPipelineStream() async* {
    while (true) {
      await Future.delayed(Duration(seconds: 3));
      if (_gitlabService.accessToken == null ||
          gitlabProject['id'] == null ||
          !projectHasPackage) {
        continue;
      } else {
        var pipelines =
            await _gitlabService.getPipelineJobs(gitlabProject['id']!);
        yield pipelines;
      }
    }
  }

  // Future<Item?> testGetOauthItem() async {
  //   var query = '''
  //     query {
  //             OauthFlow (limit: 500) {
  //               accessToken
  //               refreshToken
  //               dateCreated
  //             }
  //           }
  //     ''';

  //   var items = await _podService.graphql(query: query);
  //   if (items.length > 0) {
  //     return items[0];
  //   }
  //   return null;
  // }

  // testGitlabStuff() async {
  //   // empty repo
  //   var proj_name = "Plugin Uj4 1";

  //   // repo with model
  //   // var proj_name = "Ml Test 4";

  //   if (gitlabOauthItem == null) {
  //     gitlabOauthItem = await testGetOauthItem();
  //   }

  //   if (gitlabOauthItem == null) {
  //     AppLogger.debug('no oauth item");
  //     return null;
  //   }

  //   var accessToken = gitlabOauthItem?.get("accessToken");
  //   var refreshToken = gitlabOauthItem?.get("refreshToken");

  //   if (accessToken == null) {
  //     AppLogger.debug('no accessToken");
  //     return null;
  //   }
  //   var gitlabHelper =
  //       GitlabHelper(accessToken: accessToken, refreshToken: refreshToken);

  //   var user = await gitlabHelper.getUser();
  //   AppLogger.debug('user: $user");
  //   var userId = user["id"];
  //   AppLogger.debug('user id $userId");

  //   var projId = await gitlabHelper.getUserProjectId(userId, proj_name);
  //   AppLogger.debug('proj id: $projId");

  //   var hasPackages = await gitlabHelper.projectHasModelPackage(proj_name);
  //   AppLogger.debug('hasPackages $hasPackages");

  //   var ciStatus = await gitlabHelper.getPipelineJobStatus(projId!);
  //   AppLogger.debug('CI status $ciStatus");
  // }
  updateSelectedFeature(String name) {
    if (selectedFeatures.contains(name)) {
      selectedFeatures.removeWhere((item) => item == name);
    } else {
      selectedFeatures.add(name);
    }
    notifyListeners();
  }

  void handleSubmitButton(BuildContext context) {
    var projectId = currentProject?.get('id');
    if (datasetSummaries.containsKey(projectId)) {
      DatasetSummary summaries = datasetSummaries[projectId]!;
      AnalyticsService().logProjectLabelsOverview(
        summaries.numLabeled.toString(),
        summaries.numSkipped.toString(),
        getDatasetSize().toString(),
      );
    }
    RouteNavigator.navigateTo(
      context: context,
      route: Routes.projectsAppCreate,
      param: {'id': projectId},
    );
  }
}

class DatasetSummary {
  // TODO needs a refactor
  Item? dataset;
  int datasetSize = 0;
  int numLabeled = 0;
  int percentageLabeled = 0;
  int numSkipped = 0;
  int percentageSkipped = 0;
  int firstUnlabelledIdx = 0;
  Map<String, int> labelFrequencies = {};

  DatasetSummary({this.dataset}) {
    update();
  }

  void reset() {
    numLabeled = 0;
    datasetSize = 0;
    percentageLabeled = 0;
    numSkipped = 0;
    percentageSkipped = 0;
    firstUnlabelledIdx = 0;
    labelFrequencies = {};
  }

  void update() {
    reset();

    if (dataset != null) {
      var entries = dataset!.getEdgeTargets('entry');
      if (entries == null || entries.isEmpty) {
        return;
      }

      datasetSize = entries.length;

      labelFrequencies = {};
      for (int i = 0; i < datasetSize; i++) {
        var entry = entries[i];

        var annotations = entry.getEdgeTargets('annotation');
        if (annotations != null && annotations.isNotEmpty) {
          // user labelled item
          numLabeled++;
          String labelValue = annotations.first.get('labelValue');
          labelFrequencies[labelValue] =
              (labelFrequencies[labelValue] ?? 0) + 1;
        } else if (entry.get('skippedByLabeller') == true) {
          // user has skipped item
          numSkipped++;
        } else {
          // first time item is seen
          if (firstUnlabelledIdx == 0) {
            firstUnlabelledIdx = i;
          }
        }
      }

      if (datasetSize == 0) {
        percentageLabeled = 100;
        percentageSkipped = 100;
      } else {
        percentageLabeled =
            (numLabeled.toDouble() / datasetSize.toDouble() * 100).round();
        percentageSkipped =
            (numSkipped.toDouble() / datasetSize.toDouble() * 100).round();
      }
    }
  }

  int getLabelPercentage(int? labelFreq) {
    if (numLabeled == 0 || labelFreq == null) {
      return 0;
    } else {
      return (labelFreq.toDouble() / numLabeled.toDouble() * 100).round();
    }
  }
}
