import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_styles.dart';
import 'package:memri/core/services/analytics_service.dart';
import 'package:memri/providers/importer_provider.dart';
import 'package:memri/providers/navigation_provider.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/responsive_helper.dart';
import 'package:memri/widgets/empty.dart';
import 'package:memri/widgets/importer/importer_tile.dart';
import 'package:memri/widgets/scaffold/workspace_scaffold.dart';
import 'package:provider/provider.dart';

class DataScreen extends StatefulWidget {
  final bool showMainNavigation;

  const DataScreen({Key? key, this.showMainNavigation = true})
      : super(key: key);

  @override
  DataScreenState createState() => DataScreenState();
}

class DataScreenState extends State<DataScreen> {
  late final _navigationProvider =
      Provider.of<NavigationProvider>(context, listen: false);
  late final importerProvider =
      Provider.of<ImporterProvider>(context, listen: false);

  ScrollController _bodyScrollController = ScrollController();
  final ScrollController _sideBarScrollController = ScrollController();

  @override
  Widget build(BuildContext context) {
    return WorkspaceScaffold(
      currentItem: NavigationItem.data,
      child: Consumer<ImporterProvider>(builder: (context, provider, _) {
        return ResponsiveHelper(context).isSmallScreen
            ? _buildMobileBody(provider)
            : _buildDesktopBody(provider);
      }),
    );
  }

  Widget _buildDesktopBody(ImporterProvider provider) {
    return Row(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: [
        Expanded(
          flex: ResponsiveHelper(context).isMediumScreen ? 5 : 3,
          child: Container(
            color: app.colors.white,
            child: SingleChildScrollView(
              controller: _bodyScrollController,
              child: _buildBody(provider),
            ),
          ),
        ),
        if (provider.selectedPlugin != null) SizedBox(width: 2),
        if (provider.selectedPlugin != null)
          Expanded(
            flex: ResponsiveHelper(context).isMediumScreen ? 3 : 1,
            child: SingleChildScrollView(
              controller: _sideBarScrollController,
              child: Container(
                color: app.colors.white,
                padding: EdgeInsets.all(20),
                child: _buildSideBar(provider),
              ),
            ),
          ),
      ],
    );
  }

  Widget _buildMobileBody(ImporterProvider provider) {
    _bodyScrollController = ScrollController();
    return SingleChildScrollView(
      controller: _bodyScrollController,
      child: Column(
        children: [
          _buildBody(provider),
          if (provider.selectedPlugin != null) SizedBox(height: 2),
          if (provider.selectedPlugin != null)
            Container(
              color: app.colors.white,
              padding: EdgeInsets.all(30),
              child: _buildSideBar(provider),
            ),
        ],
      ),
    );
  }

  _buildBody(ImporterProvider provider) {
    return Container(
      width: MediaQuery.of(context).size.width,
      color: app.colors.white,
      padding: EdgeInsets.symmetric(horizontal: 30),
      child: Row(
        children: [
          Expanded(
            flex: ResponsiveHelper(context).isMediumScreen ? 5 : 3,
            child: Column(
              crossAxisAlignment: CrossAxisAlignment.start,
              children: [
                SizedBox(height: 20),
                if (!ResponsiveHelper(context).isSmallScreen)
                  SizedBox(height: 20),
                Row(
                  children: [
                    Text('Data streams', style: AppStyles.headline2),
                    Spacer(flex: 5),
                    TextButton(
                      style: ButtonStyle(
                          padding:
                              MaterialStateProperty.all(EdgeInsets.all(0))),
                      onPressed: _navigationProvider.openDataExplorerLink,
                      child: Text(
                        'Open data explorer',
                        style: AppStyles.buttonLabel
                            .copyWith(color: Color(0xffFE570F)),
                      ),
                    ),
                  ],
                ),
                SizedBox(height: 30),
                Wrap(
                  spacing: 30,
                  runSpacing: 30,
                  children: [
                    ImporterTile(
                      title: 'WhatsApp',
                      status: provider.isImporterActive(DataSource.whatsapp)
                          ? 'Running'
                          : 'Connect',
                      statusColor:
                          provider.isImporterActive(DataSource.whatsapp)
                              ? app.colors.brandGreen
                              : null,
                      active: provider.isImporterActive(DataSource.whatsapp),
                      selected:
                          provider.isImporterSelected(DataSource.whatsapp),
                      onTap: () {
                        AnalyticsService().logImporterSelect('whatsapp');
                        if (!provider.isImporterActive(DataSource.whatsapp)) {
                          RouteNavigator.navigateTo(
                            context: context,
                            route: Routes.importerConnect,
                            param: {'name': 'whatsapp'},
                          );
                        } else {
                          provider.setActivePlugin(
                              provider.activePlugins[DataSource.whatsapp]);
                        }
                      },
                    ),
                    ImporterTile(
                      title: 'Twitter',
                      status: provider.isImporterActive(DataSource.twitter)
                          ? 'Running'
                          : 'Connect',
                      statusColor: provider.isImporterActive(DataSource.twitter)
                          ? app.colors.brandGreen
                          : null,
                      active: provider.isImporterActive(DataSource.twitter),
                      selected: provider.isImporterSelected(DataSource.twitter),
                      onTap: () {
                        AnalyticsService().logImporterSelect('twitter');
                        if (!provider.isImporterActive(DataSource.twitter)) {
                          RouteNavigator.navigateTo(
                            context: context,
                            route: Routes.importerConnect,
                            param: {'name': 'twitter'},
                          );
                        } else {
                          provider.setActivePlugin(
                              provider.activePlugins[DataSource.twitter]);
                        }
                      },
                    ),
                    ImporterTile(
                      title: 'Create new importer',
                      status: 'Learn how to create your own importer',
                      selectable: false,
                      onTap: () => RouteNavigator.navigateTo(
                        route: Routes.importerCreate,
                        context: context,
                      ),
                    ),
                  ],
                ),
                SizedBox(height: 20),
                if (!ResponsiveHelper(context).isSmallScreen)
                  SizedBox(height: 130),
              ],
            ),
          ),
          if (!provider.isImporterSelected(DataSource.whatsapp))
            Expanded(
              flex: ResponsiveHelper(context).isMediumScreen ? 3 : 1,
              child: Empty(),
            ),
        ],
      ),
    );
  }

  _buildSideBar(ImporterProvider provider) {
    String pluginName =
        provider.selectedPlugin?.get('pluginName').replaceAll('Plugin', '') ??
            'unknown plugin';
    String status = provider.selectedPlugin?.get('status') ?? 'status unknown';
    String progress =
        '${(status == 'done' || status == 'daemon' ? 1.0 : provider.selectedPlugin?.get('progress') ?? 0.0) * 100}%';
    return Container(
      color: app.colors.white,
      padding: EdgeInsets.symmetric(horizontal: 20),
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          SizedBox(height: 20),
          if (!ResponsiveHelper(context).isSmallScreen) SizedBox(height: 20),
          Text(
            'selected'.toUpperCase(),
            style:
                AppStyles.smallCaps.copyWith(color: app.colors.brandGreyText),
          ),
          SizedBox(height: 30),
          Row(
            children: [
              Text(
                pluginName,
                style:
                    AppStyles.headline2.copyWith(color: app.colors.brandOrange),
              ),
              Spacer(),
              Column(
                crossAxisAlignment: CrossAxisAlignment.end,
                children: [
                  Text(
                    'status'.toUpperCase(),
                    style: AppStyles.smallCaps
                        .copyWith(color: app.colors.brandGreyText),
                  ),
                  Text(status, style: AppStyles.ul)
                ],
              )
            ],
          ),
          SizedBox(height: 20),
          Row(
            children: [
              Spacer(),
              Column(
                crossAxisAlignment: CrossAxisAlignment.end,
                children: [
                  Text(
                    'in the pod'.toUpperCase(),
                    style: AppStyles.smallCaps
                        .copyWith(color: app.colors.brandGreyText),
                  ),
                  Text(progress, style: AppStyles.ul)
                ],
              )
            ],
          ),
          SizedBox(height: 50),
          TextButton(
            onPressed: () => {
              RouteNavigator.navigateTo(
                  route: Routes.projectsCreate, context: context)
            },
            style: primaryButtonStyle,
            child: SizedBox(
                height: 36,
                child: Row(
                  mainAxisSize: MainAxisSize.min,
                  children: const [
                    Text('Create new project'),
                  ],
                )),
          ),
          SizedBox(height: 20),
          if (!ResponsiveHelper(context).isSmallScreen) SizedBox(height: 20),
        ],
      ),
    );
  }
}
