import 'package:flutter/material.dart';
import 'package:memri/constants/app_styles.dart';
import 'package:memri/providers/data_app_provider.dart';
import 'package:memri/providers/oauth_provider.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/responsive_helper.dart';
import 'package:memri/widgets/data_app/data_app_list.dart';
import 'package:memri/widgets/data_app/data_app_logs.dart';
import 'package:memri/widgets/project/project_progress_bar.dart';
import 'package:memri/widgets/scaffold/workspace_scaffold.dart';
import 'package:provider/provider.dart';

class ProjectsAppPreviewScreen extends StatefulWidget {
  const ProjectsAppPreviewScreen({Key? key, required this.id})
      : super(key: key);

  final String id;

  @override
  State<ProjectsAppPreviewScreen> createState() => _ProjectsAppPreviewScreen();
}

class _ProjectsAppPreviewScreen extends State<ProjectsAppPreviewScreen> {
  late final dataAppProvider =
      Provider.of<DataAppProvider>(context, listen: false);
  List<String> contents = ['content 1', 'content 2', 'content 3'];
  List<String> labels = ['positive', 'neutral', 'negative'];
  Map<String, Color> labelColors = {
    'positive': app.colors.brandViolet,
    'neutral': app.colors.brandGreyText,
    'negative': app.colors.brandOrange
  };

  // bool pluginRunComplete = true;

  @override
  void initState() {
    _setup();
    super.initState();
  }

  Future<void> _setup() async {
    dataAppProvider.setDataAppState(AppState.loading);

    var projectProvider = Provider.of<ProjectProvider>(context, listen: false);
    var oAuthProvider = Provider.of<OAuthProvider>(context, listen: false);
    await projectProvider.setCurrentProject(context, widget.id);
    await oAuthProvider.setGitlabTokens();
    await projectProvider.setGitlabProject();

    // for debugging, don't run app hardcode ID
    // var projId = 342;
    // dataAppProvider.currentAppState = AppState.showApp;
    // dataAppProvider.currentPluginMessages =
    //     await dataAppProvider.loadAppContent("_preview-$projId");

    var projId = projectProvider.gitlabProject['id'];
    dataAppProvider.runPluginFromGit(gitProjectId: projId!, changeState: true);
  }

  @override
  Widget build(BuildContext context) {
    return WorkspaceScaffold(
      currentItem: NavigationItem.projects,
      backgroundColor: app.colors.white,
      child: ProjectProgressBar(currentStep: 3, child: _buildBody()),
    );
  }

  String formatTimestamp(int? timestamp) {
    if (timestamp != null) {
      return DateTime.fromMillisecondsSinceEpoch(timestamp).toString();
    }
    return '';
  }

  Widget _buildBody() => Consumer<DataAppProvider>(
      builder: (context, dataAppProvider, _) =>
          dataAppProvider.currentAppState == AppState.loading
              ? _buildDeploying()
              : _buildPreview(dataAppProvider));

  Widget _buildPreview(DataAppProvider dataAppProvider) {
    return Consumer<ProjectProvider>(
      builder: (context, projectProvider, _) {
        return Container(
          color: app.colors.white,
          child: SingleChildScrollView(
            child: Row(
              children: [
                Expanded(
                  flex: 2,
                  child: Container(
                    padding: EdgeInsets.symmetric(horizontal: 30),
                    child: Column(
                      crossAxisAlignment: CrossAxisAlignment.start,
                      children: [
                        SizedBox(height: 20),
                        if (!ResponsiveHelper(context).isSmallScreen)
                          SizedBox(height: 20),
                        Text(
                          'Preview your app!',
                          style: AppStyles.headline1,
                        ),
                        SizedBox(height: 10),
                        Text(
                          'Your app is now deployed! Below you can see sample messages from your POD and predictions made by your model.',
                          style: AppStyles.bodyText1.copyWith(fontSize: 16),
                        ),
                        SizedBox(height: 40),
                        Row(
                          children: [
                            TextButton(
                                style: secondaryButtonStyle,
                                onPressed: () => dataAppProvider
                                    .setDataAppState(AppState.showApp),
                                child: Text('App')),
                            SizedBox(width: 10),
                            TextButton(
                                style: secondaryButtonStyle,
                                onPressed: () => dataAppProvider
                                    .setDataAppState(AppState.showLogs),
                                child: Text('Logs')),
                            Spacer(),
                            dataAppProvider.pluginIsReady()
                                ? TextButton(
                                    style: tertiaryButtonStyle,
                                    onPressed: () =>
                                        dataAppProvider.handleRerunApp(),
                                    child: Text(
                                      'Run app',
                                      style: AppStyles.buttonLabel,
                                    ),
                                  )
                                : Text(
                                    'Running...',
                                    style: AppStyles.buttonLabel
                                        .copyWith(color: app.colors.black),
                                  ),
                          ],
                        ),
                        SizedBox(height: 20),
                        if (dataAppProvider.currentAppState == AppState.showApp)
                          SingleChildScrollView(
                            child: Container(
                              width: MediaQuery.of(context).size.width,
                              constraints: BoxConstraints(minHeight: 256),
                              padding: EdgeInsets.all(30),
                              color: app.colors.greyBackGround,
                              child: DataAppList(
                                  contents:
                                      dataAppProvider.currentPluginMessages),
                            ),
                          )
                        else
                          DataAppLogs(
                            currentRunLog: dataAppProvider.currentRunLog,
                            backgroundColor: app.colors.greyBackGround,
                          ),
                        SizedBox(height: 30),
                        TextButton(
                          style: primaryButtonStyle,
                          onPressed: () => dataAppProvider.handlePublishApp(
                              context,
                              projectProvider.currentProject?.get('name') ??
                                  'Unnamed App'),
                          child: SizedBox(
                            height: 36,
                            child: Row(
                              mainAxisSize: MainAxisSize.min,
                              children: const [
                                Text('Next', style: AppStyles.buttonLabel),
                              ],
                            ),
                          ),
                        ),
                        SizedBox(height: 20),
                        if (!ResponsiveHelper(context).isSmallScreen)
                          SizedBox(height: 20),
                      ],
                    ),
                  ),
                ),
                if (!ResponsiveHelper(context).isSmallScreen)
                  Expanded(child: Column()),
              ],
            ),
          ),
        );
      },
    );
  }

  Widget _buildDeploying() {
    return Container(
      color: app.colors.white,
      padding: EdgeInsets.symmetric(horizontal: 30),
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          SizedBox(height: 20),
          if (!ResponsiveHelper(context).isSmallScreen) SizedBox(height: 20),
          Text(
            'Deploying your app...',
            style: AppStyles.headline1.copyWith(color: app.colors.brandViolet),
          ),
          SizedBox(height: 10),
          Text(
            'This may take a few moments',
            style: AppStyles.bodyText1.copyWith(fontSize: 16),
          ),
          SizedBox(height: 20),
          if (!ResponsiveHelper(context).isSmallScreen) SizedBox(height: 20),
        ],
      ),
    );
  }
}
