import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_styles.dart';
import 'package:memri/providers/importer_provider.dart';
import 'package:memri/providers/oauth_provider.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/responsive_helper.dart';
import 'package:memri/widgets/components/error_message.dart';
import 'package:memri/widgets/importer/importer_tile.dart';
import 'package:memri/widgets/project/project_progress_bar.dart';
import 'package:memri/widgets/scaffold/workspace_scaffold.dart';
import 'package:memri/widgets/simple_text_editor.dart';
import 'package:provider/provider.dart';

class ProjectsCreateScreen extends StatefulWidget {
  const ProjectsCreateScreen({Key? key, required this.isSampleProject})
      : super(key: key);

  @override
  State<ProjectsCreateScreen> createState() => _ProjectsCreateScreenState();
  final bool isSampleProject;
}

class _ProjectsCreateScreenState extends State<ProjectsCreateScreen> {
  late final oAuthProvider = Provider.of<OAuthProvider>(context, listen: false);
  final FocusNode _focusNode = FocusNode();
  final _userInputController = TextEditingController();
  final _projectNameController = TextEditingController();
  List<String> properties = ['content', 'dateSent'];
  String? gitlabApiKey;

  @override
  void initState() {
    var provider = Provider.of<ProjectProvider>(context, listen: false);
    provider.resetProjectState();
    oAuthProvider.initGitlabOAuth();
    if (widget.isSampleProject) {
      provider.handleSelectDataSource(DataSource.sampleWhatsapp, notify: false);
      Provider.of<ImporterProvider>(context, listen: false).importSampleData();
    }
    super.initState();
  }

  @override
  void dispose() {
    _userInputController.dispose();
    _projectNameController.dispose();
    oAuthProvider.resetGitlabStream();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return WorkspaceScaffold(
      currentItem: NavigationItem.projects,
      child: ProjectProgressBar(currentStep: 0, child: _buildBody()),
    );
  }

  Widget _buildBody() {
    return Consumer<ProjectProvider>(
      builder: (context, provider, _) {
        return Container(
          color: app.colors.white,
          child: SingleChildScrollView(
            child: Padding(
              padding: EdgeInsets.symmetric(horizontal: 30),
              child: Row(
                children: [
                  Expanded(
                    flex: 2,
                    child: Consumer<OAuthProvider>(builder: (_, oProvider, __) {
                      return Column(
                        crossAxisAlignment: CrossAxisAlignment.start,
                        children: [
                          SizedBox(height: 20),
                          if (!ResponsiveHelper(context).isSmallScreen)
                            SizedBox(height: 20),
                          Text('Create a new project',
                              style: AppStyles.headline1),
                          SizedBox(height: 50),
                          Text('Add a project name',
                              style: AppStyles.headline3),
                          SizedBox(height: 10),
                          Text(
                              'Please choose a name for your new project, this will also be the name of your app.',
                              style: AppStyles.bodyText1),
                          SizedBox(height: 20),
                          SimpleTextEditor(
                            controller: _projectNameController,
                            title: 'YOUR PROJECT NAME',
                            hintText: 'Choose a name for your project',
                          ),
                          SizedBox(height: 70),
                          Text('Authorise Gitlab', style: AppStyles.headline3),
                          SizedBox(height: 10),
                          if (!oProvider.gitlabOAuthExist)
                            Text(
                              'To be able to use your app on the data saved in your POD, you need to authorise your Memri GitLab account. ',
                              style: AppStyles.bodyText1,
                            ),
                          if (!oProvider.gitlabOAuthExist) SizedBox(height: 30),
                          if (!oProvider.gitlabOAuthExist)
                            Container(
                              decoration: BoxDecoration(
                                  border: Border(
                                      left: BorderSide(
                                          width: 3.0,
                                          color: Color(0xffDFDEDE)))),
                              padding: EdgeInsets.all(8),
                              child: Text(
                                  'Please note that gitlab.memri.io is a self-hosted gitlab, so this wont work with an account from gitlab.com.',
                                  style: AppStyles.bodyText1),
                            ),
                          if (!oProvider.gitlabOAuthExist) SizedBox(height: 30),
                          !oProvider.gitlabOAuthExist
                              ? TextButton(
                                  onPressed: oProvider.openOAuthWindow,
                                  style: tertiaryButtonStyle,
                                  child: Text(
                                    'Authorise Gitlab Access',
                                    style: AppStyles.buttonLabel,
                                  ),
                                )
                              : Row(
                                  children: [
                                    Text(
                                      'Gitlab Authorised',
                                      style: AppStyles.buttonLabel.copyWith(
                                          color: app.colors.brandGreen),
                                    ),
                                    SizedBox(width: 8),
                                    app.icons.checkmark(
                                        color: app.colors.brandGreen),
                                  ],
                                ),
                          SizedBox(height: 60),
                          Text('Select data source',
                              style: AppStyles.headline3),
                          SizedBox(height: 10),
                          Text(
                            'Select the data source for your new app. You may only select one.',
                            style: AppStyles.bodyText1,
                          ),
                          SizedBox(height: 20),
                          Consumer<ImporterProvider>(
                              builder: (_, iProvider, __) {
                            return Wrap(
                              spacing: 16,
                              runSpacing: 16,
                              children: [
                                if (iProvider.isImporterActive(
                                        DataSource.whatsapp) &&
                                    !widget.isSampleProject)
                                  ImporterTile(
                                    title: 'WhatsApp',
                                    status: 'Running',
                                    statusColor: iProvider.isImporterActive(
                                            DataSource.whatsapp)
                                        ? app.colors.brandGreen
                                        : null,
                                    selected: provider.dataSourceSelected ==
                                        DataSource.whatsapp,
                                    onTap: () =>
                                        provider.handleSelectDataSource(
                                            DataSource.whatsapp),
                                  ),
                                if (iProvider
                                        .isImporterActive(DataSource.twitter) &&
                                    !widget.isSampleProject)
                                  ImporterTile(
                                    title: 'Twitter',
                                    status: 'Running',
                                    statusColor: iProvider.isImporterActive(
                                            DataSource.twitter)
                                        ? app.colors.brandGreen
                                        : null,
                                    selected: provider.dataSourceSelected ==
                                        DataSource.twitter,
                                    onTap: () =>
                                        provider.handleSelectDataSource(
                                            DataSource.twitter),
                                  ),
                                if (widget.isSampleProject)
                                  ImporterTile(
                                    title: 'Sample Data',
                                    description:
                                        'A static dataset of 500 messages in English',
                                    selected: true,
                                    onTap: () {},
                                  ),
                              ],
                            );
                          }),
                          SizedBox(height: 60),
                          Column(
                            crossAxisAlignment: CrossAxisAlignment.start,
                            children: [
                              Text('Select feature variables',
                                  style: AppStyles.headline1),
                              SizedBox(height: 10),
                              Text(
                                  'Feature variables represent different information available in your connected datasets. You may choose the most relevant variables based on what kind of predictions you wish your app to make. For example, if you want the app use the message body, choose ‘message text’.',
                                  style: AppStyles.bodyText1),
                              SizedBox(height: 10),
                              Text(
                                  'Select which information from your dataset you wish to use in your app:',
                                  style: AppStyles.bodyText1),
                              SizedBox(height: 16),
                              Row(
                                children: [
                                  for (var feature in provider.features[
                                          provider.dataSourceSelected] ??
                                      [])
                                    _featureSelectBox(provider, feature),
                                ],
                              ),
                              SizedBox(height: 60),
                            ],
                          ),
                          Text('Add labels you want to use in your app.',
                              style: AppStyles.headline1),
                          SizedBox(height: 10),
                          Text(
                              'Add the labels you want to use on your data and in your app.  You may only use one label per data item. For example, if you are building a sentiment analysis app, add: positive, negative and neutral to label each message.',
                              style: AppStyles.bodyText1),
                          SizedBox(height: 10),
                          Text(
                              'Use the index numbers for keyboard shortcuts to help you label at warp speed!',
                              style: AppStyles.bodyText1),
                          SizedBox(height: 20),
                          Column(
                            crossAxisAlignment: CrossAxisAlignment.start,
                            children: [
                              for (var i = 0;
                                  i < provider.notSubmittedLabelValues.length;
                                  i++)
                                FittedBox(
                                  child: Container(
                                      padding: EdgeInsets.all(8),
                                      margin: EdgeInsets.symmetric(vertical: 8),
                                      decoration: BoxDecoration(
                                          color: app.colors.brandWhite,
                                          borderRadius: BorderRadius.all(
                                              Radius.circular(20))),
                                      child: Row(
                                        children: [
                                          Text(
                                            provider.notSubmittedLabelValues[i],
                                            style: AppStyles.bodyText1,
                                          ),
                                          SizedBox(width: 5),
                                          Text(
                                            (i + 1).toString(),
                                            style: AppStyles.smallCaps.copyWith(
                                                color:
                                                    app.colors.brandGreyText),
                                          ),
                                          SizedBox(width: 5),
                                          InkWell(
                                            child: app.icons.x(),
                                            onTap: () => provider
                                                .handleRemoveLabelOption(i),
                                          )
                                        ],
                                      )),
                                ),
                            ],
                          ),
                          SizedBox(height: 17),
                          Container(
                            width: 632,
                            // padding: const EdgeInsets.fromLTRB(10, 8, 10, 8),
                            color: Color(0xffF5F5F5),
                            child: Row(
                              children: [
                                Expanded(
                                  child: TextFormField(
                                    onFieldSubmitted: (str) =>
                                        provider.handleNewLabel(
                                            _userInputController, _focusNode),
                                    focusNode: _focusNode,
                                    controller: _userInputController,
                                    decoration: InputDecoration(
                                        hintText: 'e.g. positive',
                                        contentPadding: EdgeInsets.symmetric(
                                            horizontal: 10),
                                        border: InputBorder.none,
                                        hintStyle: AppStyles.input),
                                  ),
                                ),
                                TextButton(
                                  onPressed: () => provider.handleNewLabel(
                                      _userInputController, _focusNode),
                                  style: tertiaryButtonStyle,
                                  child: Text('Add new label'),
                                ),
                                SizedBox(width: 3),
                              ],
                            ),
                          ),
                          SizedBox(height: 40),
                          if (provider.errorMessage != '')
                            Padding(
                              padding: const EdgeInsets.only(bottom: 20),
                              child: ErrorMessage(provider.errorMessage),
                            ),
                          Wrap(
                            children: [
                              TextButton(
                                onPressed: () => provider.handleCreateProject(
                                    context, _projectNameController.text),
                                style: primaryButtonStyle,
                                child: SizedBox(
                                  height: 36,
                                  child: Row(
                                    mainAxisSize: MainAxisSize.min,
                                    children: [
                                      Text('Next'),
                                      SizedBox(width: 8),
                                      app.icons.arrow(),
                                    ],
                                  ),
                                ),
                              ),
                              SizedBox(width: 8),
                              TextButton(
                                onPressed: () =>
                                    RouteNavigator.navigateBack(context),
                                style: secondaryButtonStyle,
                                child: SizedBox(
                                  height: 36,
                                  child: Row(
                                    mainAxisSize: MainAxisSize.min,
                                    children: const [Text('Cancel')],
                                  ),
                                ),
                              ),
                            ],
                          ),
                          SizedBox(height: 30),
                        ],
                      );
                    }),
                  ),
                  if (ResponsiveHelper(context).isLargeScreen)
                    Expanded(child: Container()),
                ],
              ),
            ),
          ),
        );
      },
    );
  }

  Widget _featureSelectBox(ProjectProvider provider, String featureName) {
    return InkWell(
      onTap: () => provider.updateSelectedFeature(featureName),
      child: Container(
        margin: EdgeInsets.all(2),
        padding: EdgeInsets.all(8),
        decoration: BoxDecoration(
            color: provider.selectedFeatures.contains(featureName)
                ? app.colors.backgroundOrange
                : app.colors.brandWhite,
            borderRadius: BorderRadius.all(Radius.circular(2))),
        child: Row(children: [
          Text(featureName,
              style: TextStyle(
                  fontSize: 14,
                  color: provider.selectedFeatures.contains(featureName)
                      ? app.colors.brandOrange
                      : app.colors.brandGreyText)),
          app.icons.checkmark(
              color: provider.selectedFeatures.contains(featureName)
                  ? app.colors.brandOrange
                  : app.colors.brandGreyText)
        ]),
      ),
    );
  }
}
