import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_styles.dart';
import 'package:memri/core/services/analytics_service.dart';
import 'package:memri/localization/generated/l10n.dart';
import 'package:memri/providers/navigation_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/responsive_helper.dart';
import 'package:memri/widgets/navigation/menu/pane_menu.dart';
import 'package:memri/widgets/scaffold/workspace_scaffold.dart';
import 'package:provider/provider.dart';

class TopBar extends StatefulWidget implements PreferredSizeWidget {
  const TopBar({
    Key? key,
    required this.currentItem,
    this.height = kToolbarHeight,
    this.scaffoldKey,
    this.tabs,
  }) : super(key: key);

  final GlobalKey<ScaffoldState>? scaffoldKey;
  final NavigationItem currentItem;
  final double height;
  final List<TabItem>? tabs;

  @override
  State<TopBar> createState() => _TopBarState();

  @override
  Size get preferredSize => Size.fromHeight(height);
}

class _TopBarState extends State<TopBar> {
  @override
  Widget build(BuildContext context) {
    return ResponsiveHelper(context).isSmallScreen
        ? _buildMobileAppBar()
        : _buildDesktopAppBar();
  }

  Widget _buildMobileAppBar() {
    return AppBar(
      backgroundColor: app.colors.black,
      elevation: 0,
      leading: TextButton(
        onPressed: () => widget.scaffoldKey!.currentState!.openDrawer(),
        child: app.icons.hamburger(color: app.colors.white),
      ),
      centerTitle: true,
      title: InkWell(
        onTap: () {
          RouteNavigator.navigateTo(
            context: context,
            route: Routes.workspace,
            clearStack: true,
          );
        },
        child: app.images.logo(color: app.colors.white, height: 28),
      ),
    );
  }

  _buildDesktopAppBar() {
    return Consumer<NavigationProvider>(
      builder: (BuildContext context, provider, _) => Column(
        children: [
          if (!provider.isMenuVisible)
            Container(
              color: app.colors.black,
              height: 190,
              padding: EdgeInsets.fromLTRB(30, 30, 30, 0),
              child: Stack(
                children: [
                  Positioned(
                    height: 45,
                    top: 0,
                    left: 0,
                    right: 0,
                    child: Row(
                      children: [
                        TextButton(
                          onPressed: () => provider.openMenu(),
                          child: app.icons.hamburger(color: app.colors.white),
                        ),
                        SizedBox(width: 34),
                        /*SvgPicture.asset("assets/images/ico_search.svg"),*/
                        //TODO: uncomment this after search implemented
                        Spacer(),
                        TextButton(
                          onPressed: () => RouteNavigator.navigateTo(
                              context: context,
                              route: Routes.workspace,
                              clearStack: true),
                          child: Text(
                            S.current.workspace,
                            style: AppStyles.bodyText1.copyWith(
                                color: widget.currentItem ==
                                        NavigationItem.workspace
                                    ? app.colors.primary
                                    : app.colors.white),
                          ),
                        ),
                        SizedBox(width: 30),
                        TextButton(
                          onPressed: () => RouteNavigator.navigateTo(
                              context: context, route: Routes.data),
                          child: Text(
                            S.current.data,
                            style: AppStyles.bodyText1.copyWith(
                                color: widget.currentItem == NavigationItem.data
                                    ? app.colors.primary
                                    : app.colors.white),
                          ),
                        ),
                        SizedBox(width: 30),
                        TextButton(
                          onPressed: () => RouteNavigator.navigateTo(
                              context: context, route: Routes.projects),
                          child: Text(
                            S.current.projects,
                            style: AppStyles.bodyText1.copyWith(
                                color: widget.currentItem ==
                                        NavigationItem.projects
                                    ? app.colors.primary
                                    : app.colors.white),
                          ),
                        ),
                        SizedBox(width: 30),
                        TextButton(
                          onPressed: () => RouteNavigator.navigateTo(
                              context: context, route: Routes.apps),
                          child: Text(
                            S.current.apps,
                            style: AppStyles.bodyText1.copyWith(
                                color: widget.currentItem == NavigationItem.apps
                                    ? app.colors.primary
                                    : app.colors.white),
                          ),
                        ),
                        Spacer(),
                        InkWell(
                          onTap: () => RouteNavigator.navigateTo(
                            context: context,
                            route: Routes.workspace,
                            clearStack: true,
                          ),
                          child: app.images.logo(color: app.colors.white),
                        )
                      ],
                    ),
                  ),
                  if (widget.tabs != null && widget.tabs!.isNotEmpty)
                    Positioned(
                      height: 50,
                      bottom: 0,
                      left: 0,
                      right: 0,
                      child: Row(
                        children: List.generate(
                          widget.tabs!.length,
                          (index) => InkWell(
                            onTap: () {
                              provider.setSelectedTabIndex(index);
                              AnalyticsService()
                                  .logTobBarTabButton(widget.tabs![index].name);
                              widget.tabs![index].onTap!();
                            },
                            child: Container(
                              color: provider.selectedTabIndex == index
                                  ? app.colors.white
                                  : app.colors.black,
                              padding: EdgeInsets.symmetric(
                                vertical: 17,
                                horizontal: 15,
                              ),
                              child: Text(
                                widget.tabs![index].name,
                                style: AppStyles.tabList.copyWith(
                                    color: provider.selectedTabIndex == index
                                        ? app.colors.brandBlack
                                        : app.colors.brandOrange),
                              ),
                            ),
                          ),
                        ),
                      ),
                    ),
                ],
              ),
            ),
          if (provider.isMenuVisible) PaneMenu()
        ],
      ),
    );
  }
}

class TabItem {
  TabItem({required this.name, this.onTap});

  final String name;
  final GestureTapCallback? onTap;
}
