import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/localization/generated/l10n.dart';
import 'package:memri/providers/app_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/responsive_helper.dart';
import 'package:memri/widgets/loading_indicator.dart';
import 'package:provider/provider.dart';

class BaseScaffold extends StatefulWidget {
  const BaseScaffold({
    Key? key,
    this.scaffoldKey,
    required this.body,
    this.topBar,
    this.drawer,
    this.bottomBar,
    this.backgroundColor,
  }) : super(key: key);

  /// The primary content of the scaffold.
  final Widget body;

  /// An app bar to display at the top of the scaffold.
  final PreferredSizeWidget? topBar;

  /// A bottom navigation bar to display at the bottom of the scaffold.
  final Widget? bottomBar;

  /// A panel displayed to the side of the [body]
  final Widget? drawer;
  final GlobalKey<ScaffoldState>? scaffoldKey;

  /// The color of the [Material] widget that underlies the entire Scaffold.
  final Color? backgroundColor;

  @override
  State<BaseScaffold> createState() => _BaseScaffoldState();
}

class _BaseScaffoldState extends State<BaseScaffold>
    with SingleTickerProviderStateMixin {
  late Animation<Color?> _animation;
  AnimationController? _animationController;
  final colors = TweenSequence([
    TweenSequenceItem(
      weight: 1.0,
      tween: ColorTween(begin: app.colors.black, end: app.colors.white),
    ),
  ]);

  @override
  void initState() {
    _animationController = AnimationController(
        duration: const Duration(milliseconds: 500), vsync: this);
    _animation = colors.animate(_animationController!)
      ..addListener(() {
        setState(() {});
      });
    _animationController!.animateTo(1);
    super.initState();
  }

  @override
  void dispose() {
    if (_animationController != null) _animationController!.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return Consumer<AppProvider>(
      builder: (context, provider, child) {
        Widget child = Container();
        switch (provider.state) {
          case AppState.init:
          case AppState.loading:
            child = Scaffold(
              backgroundColor: _animation.value,
              body: _buildSplash(provider),
            );
            break;
          case AppState.authenticating:
          case AppState.accessKeyExist:
          case AppState.success:
            child = Scaffold(
              key: widget.scaffoldKey,
              appBar: widget.topBar,
              bottomNavigationBar: widget.bottomBar,
              body: widget.body,
              backgroundColor: widget.backgroundColor ?? app.colors.white,
              drawer: widget.drawer,
              drawerEnableOpenDragGesture:
                  ResponsiveHelper(context).isSmallScreen,
            );
            break;
          case AppState.error:
            WidgetsBinding.instance!.addPostFrameCallback((_) =>
                RouteNavigator.navigateTo(
                    context: context, route: Routes.error));
            break;
          case AppState.unauthenticated:
            WidgetsBinding.instance!
                .addPostFrameCallback((_) => RouteNavigator.navigateTo(
                      context: context,
                      route: Routes.onboarding,
                      clearStack: true,
                    ));
            break;
        }
        return child;
      },
    );
  }

  Widget _buildSplash(AppProvider provider) {
    String podVersion;
    if (provider.podVersion == null) {
      podVersion = S.current.initializing;
    } else {
      podVersion = 'V.${provider.podVersion}';
    }
    return Stack(
      children: [
        LoadingIndicator(message: provider.welcomeMessage),
        Positioned(
          bottom: 16,
          left: 0,
          right: 0,
          child: Column(
            children: [
              Text('POD: $podVersion'),
              Text('APP: V.${provider.appVersion ?? 'x.x.x.x'}'),
            ],
          ),
        ),
      ],
    );
  }
}
